/*	$NetBSD$	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (c) 2000
 *	Naofumi HONDA.  All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/unistd.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <net/if_tun.h>

#define	DEBUG_VERBOSE
#define	PPPD_DEFAULT_OPTS		"nodetach nopersist nodefaultroute"
#define	ARGBUF_SIZE	1024
#define PACKET_BUFSIZE (64 * 1024)

#define	CMDFILE_DEMAND_PPP_UP 		"/etc/ppp/demand-ppp.up"
#define	CMDFILE_DEMAND_PPP_IPUP 	"/etc/ppp/demand-ppp.ip-up"
#define	CMDFILE_DEMAND_PPP_IPDOWN 	"/etc/ppp/demand-ppp.ip-down"
#define	CONFFILE_DEMAND_PPP 		"/etc/ppp/peers/demand-ppp.conf"
#define	DATAFILE_DEMAND_PPP 		"/var/tmp/demand-ppp.ip-data"
#define	LOCKFILE_DEMAND_PPP		"/var/spool/lock"

int tunfd, sigusr1, sigusr2, verbose;
pid_t mypid, pppdpid, childpid;
char *myprog;
char datafile[ARGBUF_SIZE];
char conffile[ARGBUF_SIZE];
char lockfile[ARGBUF_SIZE];

int exec_script __P((char *));
int exec_pppd __P((char *, char *));
int open_tun __P((char *, char *));
void demand_ppp_ip_up __P(());
void demand_ppp_ip_down __P(());
void demand_ppp_interrupt __P(());
void demand_ppp_noaction __P(());
int demand_ppp_rexmit __P(());
void demand_ppp_clean_up __P(());
void demand_ppp_finish __P(());
void usage __P(());

/************************************************************
 * setup the tunnel device
 ************************************************************/
int
exec_script(s)
	char *s;
{
	int rv;
	u_char buffer[ARGBUF_SIZE];

	sprintf(buffer, "%s %s %s\n", s, conffile, datafile);
	rv = system(buffer);
	if (rv != 0)
	{
		printf("%s: %s script execution error\n", myprog, s);
	}
	return rv;
}

int
open_tun(s, tcall)
	char *s, *tcall;
{
	int fd, rv, data;
	u_char buffer[ARGBUF_SIZE];
	FILE *fp;

#ifdef	DEBUG_VERBOSE
	printf("%s: try to open tunnel device\n", myprog);
#endif	/*DEBUG_VERBOSE */

	/*
	 * setup data file.
	 */
	fp = fopen(datafile, "w");
	if (fp == NULL)
	{
		perror(myprog);
		return -1;
	}
	fprintf(fp, "# This file is automatically generated.\n\n");
	fprintf(fp, "DEMAND_PPP_MYNAME=%s\n", datafile);
	fprintf(fp, "DEMAND_PPP_PID=%d\n", mypid);
	fprintf(fp, "TUNNEL_DEVICE=%s\n", s);
	fprintf(fp, "PPP_CALL=%s\n", tcall);
	if (verbose != 0)
		fprintf(fp, "DEMAND_PPP_VERBOSE=YES\n");
	fclose(fp);
		
	/*
	 * call "demand-ppp.up" script.
	 */
	if (exec_script(CMDFILE_DEMAND_PPP_UP) != 0)
		return -1;

	/*
	 * open the tunnel device and force it a non blocking mode.
	 */
	sprintf(buffer, "/dev/%s", s);
	fd = open(buffer, O_RDWR, 0);
	if (fd < 0)
	{
		perror(myprog);
		return -1;
	}

	data = 1;
	rv = ioctl(fd, FIONBIO, &data);
	if (rv < 0)
	{
		perror(myprog);
		return -1;
	}

#ifdef	DEBUG_VERBOSE
	printf("%s: open tunnel device completed\n", myprog);
#endif	/*DEBUG_VERBOSE */
	return fd;
}

/************************************************************
 * setup pppd
 ************************************************************/
#define PPPD_NARGS 256
int
exec_pppd(args, tcall)
	char *args;
	char *tcall;
{
	char *argv[PPPD_NARGS + 1];
	u_char buffer[ARGBUF_SIZE];
	int i, num, rv, status;

#ifdef	DEBUG_VERBOSE
	printf("%s: try to excute \"pppd\"\n", myprog);
#endif	/*DEBUG_VERBOSE */

	if (strcmp(tcall, "default") == 0)
	{
		sprintf(buffer, "%s ipparam %s %s",
			args, datafile, PPPD_DEFAULT_OPTS);
	}
	else
	{
		sprintf(buffer, "%s call %s ipparam %s %s",
			args, tcall, datafile, PPPD_DEFAULT_OPTS);
	}

	argv[0] = "pppd";
	for (i = 0, num = 1; num < PPPD_NARGS; )
	{
		for (; buffer[i] == ' '; i ++)
		{
			buffer[i] = 0;
			continue;
		}
		if (buffer[i] == 0)
			break;

		argv[num ++] = &buffer[i];
		for (; buffer[i] != ' ' && buffer[i] != 0; i ++)
			;
	}
	argv[num] = 0;

	if ((pppdpid = fork()) == 0)
	{
		rv = execv("/usr/sbin/pppd", argv);
		exit(rv);
	}

	do
	{
		waitpid(pppdpid, &status, 0);
	}
	while (WIFEXITED(status) == 0 && WIFSIGNALED(status) == 0);
	pppdpid = 0;
	
	return 0;
}

/************************************************************
 * signal handling
 ************************************************************/
void
child_terminate()
{

	printf("%s: signal posted. child terminated.\n", myprog);
	exit(0);
}

int
demand_ppp_rexmit()
{
	int npackets, rv, cnt;
	fd_set rsel;
	u_char packet_buffer[PACKET_BUFSIZE];

	for (;;)
	{
		FD_ZERO(&rsel);
		FD_SET(tunfd, &rsel);
		rv = select(tunfd + 1, &rsel, NULL, NULL, NULL);
		if (rv < 0)
		{
			perror(myprog);
			return EIO;
		}

		for (npackets = 0; npackets < 1024; npackets ++)
		{
			cnt = read(tunfd, packet_buffer, PACKET_BUFSIZE);
#ifdef	DEBUG_VERBOSE
			printf("%s: read %d [0x%x][0x%x][0x%x][0x%x][0x%x]\n", 
				myprog, cnt,
				(u_int)packet_buffer[0],
				(u_int)packet_buffer[1],
				(u_int)packet_buffer[2],
				(u_int)packet_buffer[3],
				(u_int)packet_buffer[4]);
#endif	/*DEBUG_VERBOSE */

			if (cnt <= 0)
			{
				if (errno == EAGAIN)
					break;
				perror(myprog);
				return EIO;
			}

			rv = write(tunfd, packet_buffer, cnt);
			if (rv != cnt)
			{
				perror(myprog);
				return EIO;
			}
		}

		printf("%s[%d]: retransmit %d packets\n",
			 myprog, mypid, npackets);
	}
	return 0;
}
	
void
demand_ppp_ip_up()
{

#ifdef	DEBUG_VERBOSE
	printf("%s: USR1 signal posted\n", myprog);
#endif	/*DEBUG_VERBOSE */

	if (sigusr1 !=0)
		return;
	sigusr1 = 1;

	if (exec_script(CMDFILE_DEMAND_PPP_IPUP) != 0)
		return;

	if ((childpid = fork()) == 0)
	{
		signal(SIGUSR1, demand_ppp_noaction);
		signal(SIGUSR2, demand_ppp_noaction);
		signal(SIGHUP, demand_ppp_noaction);
		signal(SIGTERM, demand_ppp_noaction);
		signal(SIGINT, child_terminate);
		demand_ppp_rexmit();
		exit(0);
	}
}

void
demand_ppp_ip_down()
{

#ifdef	DEBUG_VERBOSE
	printf("%s: USR2 signal posted\n", myprog);
#endif	/*DEBUG_VERBOSE */

	if (sigusr2 !=0 || sigusr1 == 0)
		return;
	sigusr2 = 1;

	(void) exec_script(CMDFILE_DEMAND_PPP_IPDOWN);
}

void
demand_ppp_interrupt()
{
	static int block;

	if (block != 0)
		return;
	block = 1;

	demand_ppp_clean_up();
	demand_ppp_finish();
	exit(0);
}

void
demand_ppp_noaction()
{
}

/************************************************************
 * misc
 ************************************************************/
void
usage()
{

	printf("%s [-hv] [-f tunnel device][-c call][-p \"pppd options\"]\n", myprog);
}

void
demand_ppp_finish()
{

	unlink(datafile);
	unlink(lockfile);
	killpg(SIGINT, getpgrp());
}

void
demand_ppp_clean_up()
{
	int status;
	int i;

	if (tunfd >= 0)
	{
		close(tunfd);
		tunfd = -1;
	}

	if (pppdpid > 0)
	{
		for (i = 3; i > 0; i --)
		{
			kill(pppdpid, SIGINT);
			sleep(i);
		}
		waitpid(pppdpid, &status, 0);
		pppdpid = 0;
	}

	if (childpid > 0)
	{
		kill(childpid, SIGINT);
		waitpid(childpid, &status, 0);
		childpid = 0;
	}

	sigusr1 = 0;
	sigusr2 = 0;
}

/************************************************************
 * main
 ************************************************************/
int
main(argc, argv)
	int argc;
	char **argv;
{
	int rv, ch, i;
	FILE *fp;
	fd_set rsel;
	struct timeval tv;
	struct stat st;
	u_char *pppargs, *tdev, *tcall;
        extern char *optarg;
        extern int optind;
        extern int optopt;
        extern int opterr;

	/* setup defaults */
	pppargs = "/dev/tty00 115200";
	tdev = "tun0";
	tcall = "default";
	verbose = 0;

	myprog = argv[0];
	while ((ch = getopt(argc, argv, "p:f:c:hv")) != -1)
        {
                switch (ch)
                {
		case 'p':
			pppargs = optarg;
			break;
		case 'f':
			tdev = optarg;
			break;
		case 'c':
			tcall = optarg;
			break;
		case 'v':
			verbose = 1;
			break;
                case 'h':
                default:
                        usage();
			exit(1);
			break;
                }
        }

	argc -= optind;
	argv += optind;

	/* check a configuration file */
	sprintf(conffile, "%s.%s", CONFFILE_DEMAND_PPP, tcall);
	if (stat(conffile, &st) != 0)
	{
		printf("%s: %s not found!\n", myprog, conffile);
		exit(1);
	}

	/* check a lock file */
	sprintf(lockfile, "%s/demand_ppp.lock.%s", LOCKFILE_DEMAND_PPP, tdev);
	if (stat(lockfile, &st) == 0)
	{
		printf("%s: %s lock file exists!\n", myprog, lockfile);
		exit(1);
	}

	if ((fp = fopen(lockfile, "w+")) == NULL)
	{
		perror(myprog);
		exit(1);
	}
	fclose(fp);

 	if (fork() == 0) 
	{

		chdir("/");
		if (verbose == 0)
		{
			int i, nullfd;

			for (i = 0; i < FOPEN_MAX; i ++)
				close(i);
			setsid();
			
			nullfd = open("/dev/null", O_RDWR, 0);
			dup(0);
			dup(1);
		}

		signal(SIGUSR1, demand_ppp_ip_up);
		signal(SIGUSR2, demand_ppp_ip_down);
		signal(SIGINT, demand_ppp_interrupt);
		signal(SIGHUP, demand_ppp_interrupt);
		signal(SIGTERM, demand_ppp_interrupt);

		bzero(&tv, sizeof(tv));
		sigusr1 = 0;
		sigusr2 = 0;
		mypid = getpid();
		sprintf(datafile, "%s.%d",
			DATAFILE_DEMAND_PPP, mypid);
		
		for (;;)
		{
			tunfd = open_tun(tdev, tcall);
			if (tunfd < 0)
			{
				perror(myprog);
				goto bad;
			}

			FD_ZERO(&rsel);
			FD_SET(tunfd, &rsel);
			rv = select(tunfd + 1, &rsel, NULL, NULL, NULL);
			if (rv < 0)
			{
				perror(myprog);
				goto bad;
			}

			exec_pppd(pppargs, tcall);
			demand_ppp_clean_up();
			sleep(5);
		}
	}
	exit(0);

bad:
	demand_ppp_clean_up();
	demand_ppp_finish();
	exit(1);
}
