/*	$NecBSD: ct_isa.c,v 1.6.12.5 2001/06/21 04:07:38 honda Exp $	*/
/*	$NetBSD$	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1995, 1996, 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define	SCSIBUS_RESCAN
#include "opt_ct.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/disklabel.h>
#include <sys/buf.h>
#include <sys/queue.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/errno.h>

#include <vm/vm.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/scsipi/scsi_all.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsiconf.h>
#include <dev/scsipi/scsi_disk.h>

#include <dev/isa/isareg.h>
#include <dev/isa/isavar.h>
#include <dev/isa/isadmavar.h>

#include <machine/dvcfg.h>
#include <machine/physio_proc.h>
#include <machine/syspmgr.h>

#include <i386/Cbus/dev/scsi_low.h>

#include <dev/ic/wd33c93reg.h>
#include <i386/Cbus/dev/ct/ctvar.h>
#include <i386/Cbus/dev/ct/bshwvar.h>

#define	BSHW_IOSZ	0x08
#define	BSHW_IOBASE 	0xcc0
#define	BSHW_MEMSZ	(NBPG * 2)

static int ct_isa_match __P((struct device *, struct cfdata *, void *));
static void ct_isa_attach __P((struct device *, struct device *, void *));
static int ct_space_map __P((struct isa_attach_args *, struct bshw *, struct ct_bus_access_handle *));
static struct bshw *ct_find_hw __P((struct isa_attach_args *));
static void ct_isa_bus_access_weight __P((struct ct_bus_access_handle *));
static void ct_isa_dmasync_before __P((struct ct_softc *));
static void ct_isa_dmasync_after __P((struct ct_softc *));

struct ct_isa_softc {
	struct ct_softc sc_ct;
	struct bshw_softc sc_bshw;
};

struct cfattach ct_isa_ca = {
	sizeof(struct ct_isa_softc), ct_isa_match, ct_isa_attach
};

static int
ct_isa_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct isa_attach_args *ia = aux;
	struct ct_bus_access_handle ch;
	struct bshw *hw;
	int rv = 0;

	if (ia->ia_iobase == IOBASEUNK)
		ia->ia_iobase = BSHW_IOBASE;
	if ((hw = ct_find_hw(ia)) == NULL)
		return rv;
	if (ct_space_map(ia, hw, &ch) != 0)
		return rv;

	rv = ctprobesubr(&ch, 0, BSHW_DEFAULT_HOSTID, 
				 BSHW_DEFAULT_CHIPCLK, NULL);
	if (rv != 0)
	{
		struct bshw_softc bshw_tab;
		struct bshw_softc *bs = &bshw_tab;

		memset(bs, 0, sizeof(*bs));
		bshw_read_settings(&ch, bs);
		ia->ia_irq = bs->sc_irq;
		ia->ia_drq = bs->sc_drq;
		ia->ia_iosize = BSHW_IOSZ;
		if (ch.ch_memh != NULL)
			ia->ia_msize = BSHW_MEMSZ;
	}

	bus_space_unmap(ch.ch_iot, ch.ch_ioh, BSHW_IOSZ);
	if (ch.ch_memh != NULL)
		bus_space_unmap(ch.ch_memt, ch.ch_memh, BSHW_MEMSZ);
	return rv;
}

static void
ct_isa_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct ct_isa_softc *pct = (void *) self;
	struct ct_softc *ct = &pct->sc_ct;
	struct scsi_low_softc *slp = &ct->sc_sclow;
	struct ct_bus_access_handle *chp = &ct->sc_ch;
	struct bshw_softc *bs = &pct->sc_bshw;
	struct isa_attach_args *ia = aux;
	bus_dma_segment_t seg;
	struct bshw *hw;
	int rseg, error, chiprev;
	caddr_t addr;
	char *s;

	printf("\n");
	hw = ct_find_hw(ia);
	if (ct_space_map(ia, hw, chp) != 0)
		panic("%s: bus io mem map failed\n", slp->sl_dev.dv_xname);
	
	ct->sc_dmat = ia->ia_dmat;
	if (ctprobesubr(chp, 0, BSHW_DEFAULT_HOSTID, BSHW_DEFAULT_CHIPCLK,
			&chiprev) == 0)
	{
		printf("%s: hardware missing\n", slp->sl_dev.dv_xname);
		return;
	}

	/* setup DMA map */
	if (isa_dmamap_create(ia->ia_ic, ia->ia_drq, MAXBSIZE, BUS_DMA_NOWAIT))
	{
		printf("%s: can't set up ISA DMA map\n", slp->sl_dev.dv_xname);
		return;
	}

	error = bus_dmamem_alloc(ct->sc_dmat, MAXBSIZE, NBPG, 0, &seg, 1,
				 &rseg, BUS_DMA_NOWAIT);
	if (error != 0 || rseg != 1)
	{
		printf("%s: can't set up ISA DMA map\n", slp->sl_dev.dv_xname);
		return;
	}

	error = bus_dmamem_map(ct->sc_dmat, &seg, rseg, MAXBSIZE, &addr,
			       BUS_DMA_NOWAIT);
	if (error != 0)
	{
		printf("%s: can't allocate DMA mem\n", slp->sl_dev.dv_xname);
		return;
	}

	/* setup machdep softc */
	bs->sc_hw = hw;
	bs->sc_io_control = 0;
	bs->sc_bounce_phys = (caddr_t) vtophys(addr);
	bs->sc_bounce_addr = addr;
	bs->sc_bounce_size = MAXBSIZE;
	bs->sc_minphys = (1 << 24);
	bs->sc_dmasync_before = ct_isa_dmasync_before;
	bs->sc_dmasync_after = ct_isa_dmasync_after;
	bshw_read_settings(chp, bs);

	/* setup ct driver softc */
	ct->ct_hw = bs;
	ct->ct_dma_xfer_start = bshw_dma_xfer_start;
	ct->ct_pio_xfer_start = bshw_smit_xfer_start;
	ct->ct_dma_xfer_stop = bshw_dma_xfer_stop;
	ct->ct_pio_xfer_stop = bshw_smit_xfer_stop;
	ct->ct_bus_reset = bshw_bus_reset;
	ct->ct_synch_setup = bshw_synch_setup;

	ct->sc_xmode = CT_XMODE_DMA;
	if (chp->ch_memh != NULL)
		ct->sc_xmode |= CT_XMODE_PIO;

	ct->sc_chiprev = chiprev;
	switch (chiprev)
	{
	case CT_WD33C93:
		s = "WD33C93";
		ct->sc_chipclk = 8;
		break;

	case CT_WD33C93_A:
		if (DVCFG_MAJOR(ia->ia_cfgflags) > 0)
		{
			s = "AM33C93_A";
			ct->sc_chiprev = CT_AM33C93_A;
			ct->sc_chipclk = 20;
		}
		else
		{
			s = "WD33C93_A";
			ct->sc_chipclk = 10;
		}
		break;

	case CT_AM33C93_A:
		ct->sc_chipclk = 20;
		s = "AM33C93_A";
		break;

	default:
	case CT_WD33C93_B:
		s = "WD33C93_B";
		ct->sc_chipclk = 20;
		break;
	}
	printf("%s: chiprev %s chipclk %d Mhz\n", 
		slp->sl_dev.dv_xname, s, ct->sc_chipclk);

	slp->sl_hostid = bs->sc_hostid;
	slp->sl_irq = ia->ia_irq;
	slp->sl_cfgflags = ia->ia_cfgflags;

	ctattachsubr(ct);

	ct->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE, 
					IPL_BIO, ctintr, ct);

	config_found(self, slp->sl_si.si_splp, ctprint);
}

static struct bshw *
ct_find_hw(ia)
	struct isa_attach_args *ia;
{

	return DVCFG_HW(&bshw_hwsel, DVCFG_MAJOR(ia->ia_cfgflags));
}

static int
ct_space_map(ia, hw, chp)
	struct isa_attach_args *ia;
	struct bshw *hw;
	struct ct_bus_access_handle *chp;
{
	int error;

	bzero(chp, sizeof(*chp));
	chp->ch_iot = ia->ia_iot;
	chp->ch_memt = ia->ia_memt;
	chp->ch_bus_weight = ct_isa_bus_access_weight;
	syspmgr_alloc_delayh(&chp->ch_delayt, &chp->ch_delaybah);

	error = bus_space_map(chp->ch_iot, ia->ia_iobase,
			      BSHW_IOSZ, 0, &chp->ch_ioh);
	if (error != 0)
		return error;

	if ((hw->hw_flags & BSHW_SMFIFO) == 0 || ia->ia_maddr == MADDRUNK)
		return error;

	error = bus_space_map(chp->ch_memt, ia->ia_maddr,
			      BSHW_MEMSZ, 0, &chp->ch_memh);
	if (error != 0)
		bus_space_unmap(chp->ch_iot, chp->ch_ioh, BSHW_IOSZ);
	return error;
}

static void
ct_isa_bus_access_weight(chp)
	struct ct_bus_access_handle *chp;
{

	bus_space_write_1(chp->ch_delayt, chp->ch_delaybah, 0, 0);
}

#include <arch/i386/isa/pc98spec.h>	/* XXX: DMA sync: delete */

static void
ct_isa_dmasync_before(ct)
	struct ct_softc *ct;
{
	struct scsi_low_softc *slp = &ct->sc_sclow;

	if (slp->sl_scp.scp_direction == SCSI_LOW_READ)
		cpu_cf_preRead(curcpu);
	else
		cpu_cf_preWrite(curcpu);
}

static void
ct_isa_dmasync_after(ct)
	struct ct_softc *ct;
{
	struct scsi_low_softc *slp = &ct->sc_sclow;

	if (slp->sl_scp.scp_direction == SCSI_LOW_READ)
		cpu_cf_postRead(curcpu);
	else
		cpu_cf_postWrite(curcpu);
}
