/*	$NecBSD: lpt.c,v 1.13.12.3 2000/08/30 17:41:41 honda Exp $	*/
/*	$NetBSD: lpt.c,v 1.43 1996/12/05 01:25:42 cgd Exp $	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1994, 1995, 1996, 1997, 1998, 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
/*
 * Copyright (c) 1993, 1994 Charles Hannum.
 * Copyright (c) 1990 William F. Jolitz, TeleMuse
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This software is a component of "386BSD" developed by 
 *	William F. Jolitz, TeleMuse.
 * 4. Neither the name of the developer nor the name "386BSD"
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS A COMPONENT OF 386BSD DEVELOPED BY WILLIAM F. JOLITZ 
 * AND IS INTENDED FOR RESEARCH AND EDUCATIONAL PURPOSES ONLY. THIS 
 * SOFTWARE SHOULD NOT BE CONSIDERED TO BE A COMMERCIAL PRODUCT. 
 * THE DEVELOPER URGES THAT USERS WHO REQUIRE A COMMERCIAL PRODUCT 
 * NOT MAKE USE OF THIS WORK.
 *
 * FOR USERS WHO WISH TO UNDERSTAND THE 386BSD SYSTEM DEVELOPED
 * BY WILLIAM F. JOLITZ, WE RECOMMEND THE USER STUDY WRITTEN 
 * REFERENCES SUCH AS THE  "PORTING UNIX TO THE 386" SERIES 
 * (BEGINNING JANUARY 1991 "DR. DOBBS JOURNAL", USA AND BEGINNING 
 * JUNE 1991 "UNIX MAGAZIN", GERMANY) BY WILLIAM F. JOLITZ AND 
 * LYNNE GREER JOLITZ, AS WELL AS OTHER BOOKS ON UNIX AND THE 
 * ON-LINE 386BSD USER MANUAL BEFORE USE. A BOOK DISCUSSING THE INTERNALS 
 * OF 386BSD ENTITLED "386BSD FROM THE INSIDE OUT" WILL BE AVAILABLE LATE 1992.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE DEVELOPER BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Device Driver for AT parallel printer port
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/device.h>
#include <sys/conf.h>
#include <sys/syslog.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/isa/isavar.h>
#include <i386/Cbus/dev/lptreg.h>
#include <i386/Cbus/dev/lpthw.h>
#include <machine/syspmgr.h>

#define	TIMEOUT		hz*16	/* wait up to 16 seconds for a ready */
#define	STEP		hz/4

#define	LPTPRI		(PZERO+8)
#define	LPT_BSIZE	NBPG

#if !defined(DEBUG) || !defined(notdef)
#define LPRINTF(a)
#else
#define LPRINTF		if (lptdebug) printf a
int lptdebug = 1;
#endif

struct lpt_softc {
	struct device sc_dev;
	void *sc_ih;

	size_t sc_count;
	struct buf *sc_inbuf;
	u_char *sc_cp;
	int sc_spinmax;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	int sc_irq;

	int sc_state;
#define	LPT_OPEN	0x01	/* device is open */
#define	LPT_OBUSY	0x02	/* printer is busy doing output */
#define	LPT_INIT	0x04	/* waiting to initialize for open */
#define	LPT_MSGWAIT	0x10	/* wait status recovery */

	u_int sc_flags;
#define	LPT_AUTOLF	0x20	/* automatic LF on CR */
#define	LPT_NOPRIME	0x40	/* don't prime on open */
#define	LPT_NOINTR	0x80	/* do not use interrupt */

	int sc_nintr;		/* interrupts count */
	int sc_nfifo;		/* fifo size */

	u_int8_t sc_control;
	u_int8_t sc_laststatus;
	u_int8_t sc_ecr;

	int sc_mode;
#define	LPT_MODE_OLD		0	/* only busy line available */
#define	LPT_MODE_STANDARD	1	/* 3 lines handshakes */
#define	LPT_MODE_PS2		2
#define	LPT_MODE_FIFO		3
#define	LPT_MODE_ECP		4
#define	DVCFG_ACCESS_MODE	7

	struct lpt_hw *sc_hw;
};

#include <i386/Cbus/dev/lpthwtab.h>
static u_char *lpt_mode_strings[] = {"old", "standard", "ps2", "fifo", "ecp",
			     	    "unk", "unk", "unk"};

/* XXX does not belong here */
cdev_decl(lpt);

int lptprobe __P((struct device *, struct cfdata *, void *));
void lptattach __P((struct device *, struct device *, void *));
int lptintr __P((void *));

struct cfattach lpt_ca = {
	sizeof(struct lpt_softc), lptprobe, lptattach
};

extern struct cfdriver lpt_cd;

#define	LPTUNIT(s)	(minor(s) & 0x1f)
#define	LPTFLAGS(s)	(minor(s) & 0xe0)

#define	LPT_FIFO_SIZE	16
#define	LPS_INVERT	(LPS_SELECT|LPS_NERR|LPS_NBSY|LPS_NACK)
#define	LPS_MASK	(LPS_SELECT|LPS_NERR|LPS_NBSY|LPS_NACK|LPS_NOPAPER)
#define	NOT_READY_ERR(sc)	((*(sc)->sc_hw->not_ready_error)((sc)))


static void lpt_watchdog __P((void *arg));
static int pushbytes __P((struct lpt_softc *, int));
static int __lptprobe __P((struct isa_attach_args *));

int	lpt_port_test __P((bus_space_tag_t, bus_space_handle_t, bus_addr_t,
	    bus_size_t, u_char, u_char));
void lpt_message __P((void *));
static u_int8_t lpt_setup_mode __P((bus_space_tag_t, bus_space_handle_t, struct lpt_hw *, u_int));
void lpt_shutdown __P((struct lpt_softc *));
static int lpt_write_data __P((struct lpt_softc *));
static int lpt_default_mode __P((struct lpt_hw *, int));

/* mode initialize */
static int
lpt_default_mode(hw, extmode)
	struct lpt_hw *hw;
	int extmode;
{

	if ((hw->flags & LPT_CAPABILITY_MODEFIXED) != 0)
	{
		extmode = hw->mode;
	}
	else if ((hw->flags & LPT_CAPABILITY_EXTREGS) == 0)
	{
		extmode = LPT_MODE_STANDARD;
	}
	else if (extmode == 0 || extmode > hw->mode)
	{
		extmode = hw->mode;
	}
	return extmode;
}

static u_int8_t
lpt_setup_mode(bst, bsh, hw, mode)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	struct lpt_hw *hw;
	u_int mode;
{
	u_int8_t regv, ecr;
	
	switch (hw->subtype)
	{
	case LPT_HW_NEC:
		regv = bus_space_read_1(bst, bsh, lpt_mode);
		bus_space_write_1(bst, bsh, lpt_mode, regv | LPTM_EXT);
		delay(100);
		break;

	case LPT_HW_TRADNEC:
		regv = bus_space_read_1(bst, bsh, lpt_mode);
		bus_space_write_1(bst, bsh, lpt_mode, regv & ~LPTM_EXT);
		delay(100);
		syspmgr(SYSTM_SYSPMGR_TAG, SYSPMGR_PSTB_OFF);
		bus_space_write_1(bst, bsh, lpt_control, LPC_MODE8255);
		bus_space_write_1(bst, bsh, lpt_control, LPC_NIRQ8);
		bus_space_write_1(bst, bsh, lpt_control, LPC_NPSTB);
		syspmgr(SYSTM_SYSPMGR_TAG, SYSPMGR_PSTB_ON);
		break;

	default:
		break;
	}

	if ((hw->flags & LPT_CAPABILITY_EXTREGS) == 0)
		return 0;

	ecr = LPTEM_SINTR | LPTEM_EINTR;
	switch (mode)
	{
	default:
		ecr |= LPTEM_STAND;
		break;
	case LPT_MODE_PS2:		
		ecr |= LPTEM_PS2;
		break;
	case LPT_MODE_FIFO:		
		ecr |= LPTEM_FIFO;
		break;
	case LPT_MODE_ECP:		
		ecr |= LPTEM_ECP;
		break;
	}

	bus_space_write_1(bst, bsh, lpt_extmode, ecr);
	return ecr;
}

/* status */
int
not_ready_error_old(sc)
	struct lpt_softc *sc;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh;

	return ((bus_space_read_1(iot, ioh, lpt_status) & LPS_O_NBSY) != LPS_O_NBSY);
}

int
not_ready_error_ps2(sc)
	struct lpt_softc *sc;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh;
	u_int8_t new, status;

	status = ((bus_space_read_1(iot, ioh, lpt_status) ^ LPS_INVERT) & LPS_MASK);
	new = (status && status != sc->sc_laststatus) ? status : 0;
	sc->sc_laststatus = status;

	if (new && (sc->sc_state & LPT_MSGWAIT) == 0)
	{
		sc->sc_state |= LPT_MSGWAIT;
		timeout(lpt_message, sc, 30 * hz);
	}
	return ((u_int)status);
}

void
lpt_shutdown(sc)
	struct lpt_softc *sc;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh;

	sc->sc_state = 0;
	switch (sc->sc_mode)
	{
	case LPT_MODE_OLD:
		break;

	default:
		bus_space_write_1(iot, ioh, lpt_control, LPC_NINIT);
		break;
	}
}

/* misc */
void
lpt_message(arg)
	void *arg;
{
	struct lpt_softc *sc = arg;
	u_int8_t status;

	status = bus_space_read_1(sc->sc_iot, sc->sc_ioh, lpt_status);
	status = ((status ^ LPS_INVERT) & LPS_MASK);
	if (status & (LPS_SELECT | LPS_NOPAPER | LPS_NERR))
	{
		printf("%s:", sc->sc_dev.dv_xname);
		if (status & LPS_SELECT)
			printf(" [off line]");
		if (status & LPS_NOPAPER)
			printf(" [out of paper]");
		if (status & LPS_NERR)
			printf(" [output error]");
		printf("\n");
	}
	sc->sc_state &= ~LPT_MSGWAIT;
}

/*
 * Internal routine to lptprobe to do port tests of one byte value.
 */
int
lpt_port_test(iot, ioh, base, off, data, mask)
	bus_space_tag_t iot;
	bus_space_handle_t ioh;
	bus_addr_t base;
	bus_size_t off;
	u_char data, mask;
{
	int timeout;
	u_char temp;

	data &= mask;
	bus_space_write_1(iot, ioh, off, data);
	timeout = 1000;
	do {
		delay(10);
		temp = bus_space_read_1(iot, ioh, off) & mask;
	} while (temp != data && --timeout);
	LPRINTF(("lpt: port=0x%x out=0x%x in=0x%x timeout=%d\n", base + off,
	    data, temp, timeout));
	return (temp == data);
}

/*
 * Logic:
 *	1) You should be able to write to and read back the same value
 *	   to the data port.  Do an alternating zeros, alternating ones,
 *	   walking zero, and walking one test to check for stuck bits.
 *
 *	2) You should be able to write to and read back the same value
 *	   to the control port lower 5 bits, the upper 3 bits are reserved
 *	   per the IBM PC technical reference manauls and different boards
 *	   do different things with them.  Do an alternating zeros, alternating
 *	   ones, walking zero, and walking one test to check for stuck bits.
 *
 *	   Some printers drag the strobe line down when the are powered off
 * 	   so this bit has been masked out of the control port test.
 *
 *	   XXX Some printers may not like a fast pulse on init or strobe, I
 *	   don't know at this point, if that becomes a problem these bits
 *	   should be turned off in the mask byte for the control port test.
 *
 *	3) Set the data and control ports to a value of 0
 */
int
lptprobe(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct isa_attach_args *ia = aux;
	struct lpt_hw *hw;
	u_long dvcfg, i;
	int rv;

	if (ia->ia_iobase == 0x40 || ia->ia_iobase == 0x140)
	{
		for (i = 2; i > 0; i --)
		{
			dvcfg = DVCFG_MKCFG(i, DVCFG_MINOR(ia->ia_cfgflags));
			hw = DVCFG_HW(&lpt_hwsel, DVCFG_MAJOR(dvcfg));
			ia->ia_iobase = hw->iobase;
			ia->ia_irq = hw->irq;
			ia->ia_cfgflags = dvcfg;
			rv = __lptprobe(ia);
			if (rv != 0)
				return rv;
		}
	}
	else
	{
		dvcfg = DVCFG_MKCFG(0, DVCFG_MINOR(ia->ia_cfgflags));
		ia->ia_cfgflags = dvcfg;
		rv = __lptprobe(ia);
		return rv;
	}
	return 0;
}


static int
__lptprobe(ia)
	struct isa_attach_args *ia;
{
	bus_space_tag_t iot;
	bus_space_handle_t ioh;
	u_long base;
	u_char mask, data;
	int i, rv = 0;
	struct lpt_hw *hw;
	u_int extmode;

	iot = ia->ia_iot;
	base = ia->ia_iobase;
	hw = DVCFG_HW(&lpt_hwsel, DVCFG_MAJOR(ia->ia_cfgflags));
	if (hw == NULL)
	{
		return rv;
	}

#ifdef	LPT_DEBUG_OLDMODE
	if (hw->subtype != LPT_HW_TRADNEC)
		return 0;
#endif	/* LPT_DEBUG_OLDMODE */

	if (bus_space_map(iot, base, 0, 0, &ioh) != 0 ||
	    bus_space_map_load(iot, ioh, hw->iatsz, hw->iat,
			       BUS_SPACE_MAP_FAILFREE) != 0)
	{
		return rv;
	}

	if (hw->subtype == LPT_HW_NEC)
	{
		/* check phantom images */
		bus_space_write_1(iot, ioh, lpt_data, 0x0);
		bus_space_write_1(iot, ioh, lpt_alt_data, 0xaa);
		if (bus_space_read_1(iot, ioh, lpt_data) == 0xaa)
			goto bad;
	}

	extmode = DVCFG_MINOR(ia->ia_cfgflags) & DVCFG_ACCESS_MODE;
	extmode = lpt_default_mode(hw, extmode);
	switch (extmode)
	{
	case LPT_MODE_OLD:
		lpt_setup_mode(iot, ioh, hw, extmode);
		break;

	default:
		lpt_setup_mode(iot, ioh, hw, LPT_MODE_STANDARD);
		rv = 0;
		mask = 0xff;

		data = 0x55;				/* Alternating zeros */
		if (!lpt_port_test(iot, ioh, base, lpt_data, data, mask))
			goto bad;

		data = 0xaa;				/* Alternating ones */
		if (!lpt_port_test(iot, ioh, base, lpt_data, data, mask))
			goto bad;

		for (i = 0; i < CHAR_BIT; i++) {	/* Walking zero */
			data = ~(1 << i);
			if (!lpt_port_test(iot, ioh, base, lpt_data, data, mask))
				goto bad;
		}

		for (i = 0; i < CHAR_BIT; i++) {	/* Walking one */
			data = (1 << i);
			if (!lpt_port_test(iot, ioh, base, lpt_data, data, mask))
				goto bad;
		}

		bus_space_write_1(iot, ioh, lpt_data, 0);
		bus_space_write_1(iot, ioh, lpt_control, 0);
	}

	ia->ia_iosize = hw->nports;
	ia->ia_msize = 0;
	rv = 1;

bad:
	bus_space_unmap(iot, ioh, hw->nports);
	return rv;
}

void
lptattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct lpt_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;
	bus_space_tag_t iot;
	bus_space_handle_t ioh;
	u_int extmode;

	sc->sc_irq = ia->ia_irq;
	sc->sc_state = 0;
	sc->sc_nfifo = LPT_FIFO_SIZE;	/* XXX */
	iot = sc->sc_iot = ia->ia_iot;
	sc->sc_hw = DVCFG_HW(&lpt_hwsel, DVCFG_MAJOR(ia->ia_cfgflags));

	if (bus_space_map(iot, ia->ia_iobase, 0, 0, &ioh) ||
	    bus_space_map_load(iot, ioh, sc->sc_hw->iatsz, sc->sc_hw->iat,
			       BUS_SPACE_MAP_FAILFREE))
	{
		printf("%s: can not map\n", sc->sc_dev.dv_xname);
		return;
	}

	sc->sc_ioh = ioh;
	extmode = DVCFG_MINOR(ia->ia_cfgflags) & DVCFG_ACCESS_MODE;
	extmode = lpt_default_mode(sc->sc_hw, extmode);
	sc->sc_mode = extmode;
	sc->sc_ecr = lpt_setup_mode(iot, ioh, sc->sc_hw, extmode);
	lpt_shutdown(sc);

	printf(": %s mode", lpt_mode_strings[extmode]);
	if (ia->ia_irq != IRQUNK)
		printf("\n");
	else
		printf(", polled\n");
	
	if (ia->ia_irq != IRQUNK)
	{
		sc->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE,
		    IPL_TTY, lptintr, sc);
	}
}

/*
 * Reset the printer, then wait until it's selected and not busy.
 */
int
lptopen(dev, flag, mode, p)
	dev_t dev;
	int flag;
	int mode;
	struct proc *p;
{
	int unit = LPTUNIT(dev);
	u_char flags = LPTFLAGS(dev);
	struct lpt_softc *sc;
	bus_space_tag_t iot;
	bus_space_handle_t ioh;
	u_char control;
	int error;
	int spin;

	if (unit >= lpt_cd.cd_ndevs)
		return ENXIO;
	sc = lpt_cd.cd_devs[unit];
	if (!sc)
		return ENXIO;

	if (sc->sc_irq == IRQUNK && (flags & LPT_NOINTR) == 0)
		return ENXIO;

#ifdef DIAGNOSTIC
	if (sc->sc_state)
		printf("%s: stat=0x%x not zero\n", sc->sc_dev.dv_xname,
		    sc->sc_state);
#endif

	if (sc->sc_state)
		return EBUSY;
	sc->sc_state = LPT_INIT;

	iot = sc->sc_iot;
	ioh = sc->sc_ioh;
	lpt_setup_mode(iot, ioh, sc->sc_hw, sc->sc_mode);

	switch (sc->sc_mode)
	{
	case LPT_MODE_OLD:
		flags |= LPT_NOINTR;
		control = 0;
		break;

	default:
		if ((flags & LPT_NOPRIME) == 0)
		{
			bus_space_write_1(iot, ioh, lpt_control, LPC_SELECT);
			delay(100);
		}

		control = LPC_SELECT | LPC_NINIT;
		bus_space_write_1(iot, ioh, lpt_control, control);
		break;
	}
	sc->sc_flags = flags;
	sc->sc_laststatus = 0;

	/* wait till ready (printer running diagnostics) */
	for (spin = 0; NOT_READY_ERR(sc) != 0; spin += STEP) 
	{
		if (spin >= TIMEOUT)
		{
			sc->sc_state = 0;
			return EBUSY;
		}

		/* wait 1/4 second, give up if we get a signal */
		error = tsleep((caddr_t)sc, LPTPRI | PCATCH, "lptopen", STEP);
		if (error != EWOULDBLOCK)
		{
			sc->sc_state = 0;
			return error;
		}
	}

	switch (sc->sc_mode)
	{
	case LPT_MODE_OLD:
		break;

	default:
	case LPT_MODE_PS2:
	case LPT_MODE_STANDARD:
		if ((flags & LPT_NOINTR) == 0)
			control |= LPC_IENABLE;
		/* go throught */

	case LPT_MODE_FIFO:
	case LPT_MODE_ECP:
		if ((flags & LPT_AUTOLF) != 0)
			control |= LPC_AUTOLF;
		bus_space_write_1(iot, ioh, lpt_control, control);
		break;
	}

	sc->sc_control = control;
	sc->sc_laststatus = 0;
	sc->sc_inbuf = geteblk(LPT_BSIZE);
	sc->sc_count = 0;
	sc->sc_state = LPT_OPEN;

	if ((sc->sc_flags & LPT_NOINTR) == 0)
	{
		lpt_watchdog(sc);
	}

	LPRINTF(("%s: opened\n", sc->sc_dev.dv_xname));
	return 0;
}

void
lpt_watchdog(arg)
	void *arg;
{
	struct lpt_softc *sc = arg;
	int s;

	s = spltty();
	if (sc->sc_nintr == 0)
	{
		lptintr(sc);
	}
	else
	{
		sc->sc_nintr = 0;
	}
	splx(s);

	timeout(lpt_watchdog, sc, STEP);
}

/*
 * Close the device, and free the local line buffer.
 */
int
lptclose(dev, flag, mode, p)
	dev_t dev;
	int flag;
	int mode;
	struct proc *p;
{
	int unit = LPTUNIT(dev);
	struct lpt_softc *sc = lpt_cd.cd_devs[unit];
	int s;

	if (sc->sc_count > 0)
		(void) pushbytes(sc, 0);

	s = spltty();
	sc->sc_count = 0;	/* keep this */
	if ((sc->sc_flags & LPT_NOINTR) == 0)
	{
		untimeout(lpt_watchdog, sc);
	}
	untimeout(lpt_message, sc);

	lpt_setup_mode(sc->sc_iot, sc->sc_ioh, sc->sc_hw, sc->sc_mode);
	lpt_shutdown(sc);
	splx(s);

	brelse(sc->sc_inbuf);
	return 0;
}

int
pushbytes(sc, n)
	struct lpt_softc *sc;
	int n;
{
	int error;

	if ((sc->sc_flags & LPT_NOINTR) != 0)
	{
		int spin, tic;

		if (sc->sc_count == 0)
		{
			sc->sc_cp = sc->sc_inbuf->b_data;
			sc->sc_count = n;
		}

		while (sc->sc_count > 0)
		{
			spin = 0;
			while (lpt_write_data(sc) == 0) 
			{
				if (++spin < sc->sc_spinmax)
					continue;
				tic = 0;
				/* adapt busy-wait algorithm */
				sc->sc_spinmax++;
				while (NOT_READY_ERR(sc) != 0)
				{
					/* exponential backoff */
					tic = tic + tic + 1;
					if (tic > TIMEOUT)
						tic = TIMEOUT;
					error = tsleep((caddr_t)sc,
					    LPTPRI | PCATCH, "lptpsh", tic);
					if (error != EWOULDBLOCK)
						return error;
				}
				break;
			}

			/* adapt busy-wait algorithm */
			if (spin*2 + 16 < sc->sc_spinmax)
				sc->sc_spinmax--;
		}
	}
	else 
	{
		int s;

		s = spltty();
		if (sc->sc_count == 0)
		{
			sc->sc_cp = sc->sc_inbuf->b_data;
			sc->sc_count = n;
		}

		while (sc->sc_count > 0)
		{
			if ((sc->sc_state & LPT_OBUSY) == 0)
			{
				(void) lptintr(sc);
			}
			error = tsleep((caddr_t)sc, LPTPRI | PCATCH,
			    "lptwrite2", 0);
			if (error)
			{
				splx(s);
				return error;
			}
		}
		splx(s);
	}
	return 0;
}

/* 
 * Copy a line from user space to a local buffer, then call putc to get the
 * chars moved to the output queue.
 */
int
lptwrite(dev, uio, flags)
	dev_t dev;
	struct uio *uio;
	int flags;
{
	struct lpt_softc *sc = lpt_cd.cd_devs[LPTUNIT(dev)];
	size_t n;
	int error = 0, s;

	while ((n = min(LPT_BSIZE, uio->uio_resid)) != 0)
	{
		uiomove(sc->sc_inbuf->b_data, n, uio);
		error = pushbytes(sc, n);
		if (error != 0)
		{
			/*
			 * Return accurate residual if interrupted or timed
			 * out.
			 */
			s = spltty();
			uio->uio_resid += sc->sc_count;
			sc->sc_count = 0;
			splx(s);
			return error;
		}
	}
	return 0;
}

/*
 * Handle printer interrupts which occur when the printer is ready to accept
 * another char.
 */
static int
lpt_write_data(sc)
	struct lpt_softc *sc;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_ioh;
	u_int8_t regv;

	switch (sc->sc_mode)
	{
	case LPT_MODE_OLD:
		regv = bus_space_read_1(iot, ioh, lpt_status);
		if ((regv & LPS_O_NBSY) != LPS_O_NBSY)
			return 0;

		if (sc->sc_count > 0)
		{
			bus_space_write_1(iot, ioh, lpt_data, *sc->sc_cp ++);
			bus_space_write_1(iot, ioh, lpt_control, LPC_PSTB);
			sc->sc_count--;
			sc->sc_laststatus = 0;
			bus_space_write_1(iot, ioh, lpt_control, LPC_NPSTB);
		}
		break;

	default:
	case LPT_MODE_STANDARD:
	case LPT_MODE_PS2:
		regv = bus_space_read_1(iot, ioh, lpt_status);
		regv = (regv ^ LPS_INVERT) & LPS_MASK;
		if (regv != 0)
			return 0;

		if (sc->sc_count > 0)
		{
			bus_space_write_1(iot, ioh, lpt_data, *sc->sc_cp++);
			bus_space_write_1(iot, ioh, lpt_control,
					  sc->sc_control | LPC_STROBE);
			sc->sc_count--;
			sc->sc_laststatus = 0;
			bus_space_write_1(iot, ioh, lpt_control, sc->sc_control);
		}
		break;

	case LPT_MODE_FIFO:
	case LPT_MODE_ECP:
		regv = bus_space_read_1(iot, ioh, lpt_status);
#if	0
		regv = ((regv ^ LPS_INVERT) & (LPS_MASK & ~LPS_NBSY)) ;
#else
		regv = ((regv ^ LPS_INVERT) & LPS_MASK) ;
#endif
		if (regv != 0)
		{
			return 0;
		}

		regv = bus_space_read_1(iot, ioh, lpt_extmode);
		if ((regv & LPTEM_FIFOEMP) == 0)
		{
			return (sc->sc_flags & LPT_NOINTR) ? 1 : 0;
		}

		if (sc->sc_count > 0)
		{
			int n;

			n = sc->sc_nfifo;
			if (n > sc->sc_count)
				n = sc->sc_count;

			bus_space_write_multi_1(iot, ioh, lpt_fdata,
				sc->sc_cp, n);

			if ((sc->sc_flags & LPT_NOINTR) == 0)
				bus_space_write_1(iot, ioh, lpt_extmode, 
						  sc->sc_ecr & ~LPTEM_SINTR);
			else	
				bus_space_write_1(iot, ioh, lpt_extmode,
						  sc->sc_ecr);
			sc->sc_cp += n;
			sc->sc_count -= n;
			sc->sc_laststatus = 0;
		}
		else
		{
			if ((sc->sc_flags & LPT_NOINTR) == 0)
			{
				bus_space_write_1(iot, ioh, lpt_extmode,
						  sc->sc_ecr);
			}
		}
		break;
	}
	return 1;
}

int
lptintr(arg)
	void *arg;
{
	struct lpt_softc *sc = arg;
	int ocount;

	sc->sc_nintr ++;
	ocount = sc->sc_count;
	if (lpt_write_data(sc) != 0)
	{
		if (ocount > 0)
		{
			sc->sc_state |= LPT_OBUSY;
		}
		else
		{
			sc->sc_state &= ~LPT_OBUSY;
			wakeup((caddr_t)sc);
		}
		return 1;
	}

	(void) NOT_READY_ERR(sc);
	return 0;
}

int
lptioctl(dev, cmd, data, flag, p)
	dev_t dev;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	int error = 0;

	switch (cmd) {
	default:
		error = ENODEV;
	}

	return error;
}
