/*	$NetBSD$	*/
#ifdef	BOOT_DISK_DIRECT
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * DMA and multi-sector PIO handling are derived from code contributed by
 * Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles M. Hannum.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <sys/disk.h>
#include <sys/syslog.h>
#include <sys/errno.h>

#include "disk_direct.h"
#include "pccs.h"
#include "wdcsubr.h"
#include "wdc_pccs.h"
#include "wdc_isa.h"
#include "nsp_pccs.h"

int disk_direct_key;

static void *disk_direct_sc;
static int disk_direct_select;
static int disk_direct_dev;

struct disk_direct_op {
	int dop_no;
	int dop_dev;
	int (*dop_primary) __P((int));
	void *(*dop_attach) __P((struct disk_attach_args *));
	int (*dop_read) __P((void *, daddr_t, daddr_t, u_int8_t *));
};

static struct disk_direct_op ddop[] = {
#ifdef	BOOT_PCCS_WDC
	{ 0,
	  0x80,
	  wdc_pccs_primary,
	  wdc_pccs_attach,
	  wdc_pccs_read
	},
#endif	/* BOOT_PCCS_WDC */

#ifdef	BOOT_ISA_WDC
	{ 1,
	  0x80,
	  NULL,
	  wdc_isa_attach,
	  wdc_isa_read
	},
#endif	/* BOOT_ISA_WDC */

#ifdef	BOOT_PCCS_NSP
	{ 2,
	  0xa0,
	  NULL,
	  nsp_pccs_attach,
	  nsp_pccs_read
	},
#endif	/* BOOT_PCCS_NSP */

	{-1,
	 0,
	 NULL,
	 NULL,
	 NULL	
	}
};	

int
disk_direct_primary(biosdev)
	int biosdev;
{
	struct disk_direct_op *dop = &ddop[0];
	int res = biosdev;

 	disk_direct_key = DISK_DIRECT_KEY;

	for ( ; dop->dop_no >= 0; dop ++)
	{
#ifdef	BOOT_DISK_DIRECT_PRIMARY
		if (dop->dop_primary != NULL)
		{
			res = (*dop->dop_primary)(biosdev);
			if (res != biosdev)
				break;
		}
#endif	/* BOOT_DISK_DIRECT_PRIMARY */
	}

	return res;
}

void *
disk_direct_attach(da)
	struct disk_attach_args *da;
{
	struct disk_direct_op *dop = &ddop[0];
	void *res = NULL;

	disk_direct_sc = res;
	disk_direct_dev = 0;
	for ( ; dop->dop_no >= 0; dop ++)
	{
		if ((da->da_dev & 0xf0) != dop->dop_dev)
			continue;

		res = (*dop->dop_attach) (da);
		if (res != NULL)
		{
			disk_direct_select = dop->dop_no;
			disk_direct_dev = da->da_dev;
			disk_direct_sc = res;
			break;
		}
	}
	return res;
}

int
disk_direct_read(dev, blkno, nblks, bufp)
	int dev;
	daddr_t blkno;
	daddr_t nblks;
	u_int8_t *bufp;
{
	struct disk_direct_op *dop;
	int error = EIO;

	if (disk_direct_sc == NULL || dev != disk_direct_dev)
		return error;

 	dop = &ddop[disk_direct_select];
	if ((dev & 0xf0) != dop->dop_dev)
		return error;

	error = (*dop->dop_read) (disk_direct_sc, blkno, nblks, bufp);
	return error;
}
#endif	/* BOOT_DISK_DIRECT */
