#ifdef	BOOT_IPPI
/*	$NetBSD$	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1996, 1997, 1998, 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
/*
 * Copyright (c) 1996, Sujal M. Patel
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: pnpcfg.c,v 1.6 1996/05/06 02:08:25 smpatel Exp smpatel $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/errno.h>
#include <machine/cpu.h>
#include <lib/libsa/stand.h>

#include "busio_user.h"
#include "ippivar.h"

static void ippi_print __P((struct ippi_softc *, struct ippi_board_info *));
static void ippi_wakeup_target __P((struct ippi_softc *, struct ippi_board_info *, struct ippi_ld_info *));
static void ippi_wfk_target __P((struct ippi_softc *));
static void ippi_activate_target __P((struct ippi_softc *, int));
static int ippi_is_active __P((struct ippi_softc *));

static void ippi_regw __P((struct ippi_softc *, bus_addr_t, u_int8_t));
static u_int8_t ippi_regr __P((struct ippi_softc *, bus_addr_t));
static void ippi_regw_2 __P((struct ippi_softc *, bus_addr_t, u_int16_t));
static u_int16_t ippi_regr_2 __P((struct ippi_softc *, bus_addr_t));

static int ippi_space_prefer __P((struct ippi_softc *, ippres_res_t));
static int ippi_find_resources __P((struct ippi_softc *, u_int8_t *, u_int, slot_device_res_t));
static int ippi_space_prefer __P((struct ippi_softc *, ippres_res_t));
static int ippres_parse_items __P((u_int8_t *, struct ippres_request *, ippres_res_t));
static int ippi_get_serial __P((struct ippi_softc *sc, struct serial_id *));
static int ippi_isolation __P((struct ippi_softc *sc, int));
static int ippi_scan_board __P((struct ippi_softc *, int));
static void ippi_send_initiation_LFSR __P((struct ippi_softc *sc));
static void ippi_setup_ib __P((struct ippi_board_info *, u_int, struct serial_id *));
static void ippi_read_info __P((struct ippi_softc *, struct ippi_board_info *, struct serial_id *));
static int ippi_read_config __P((struct ippi_softc *, slot_device_res_t));

static u_char *dv_xname = "ippi";
static int ippi_ndev;

/************************************************************
 * port registers
 ************************************************************/
#define	ippi_wdofs	0x800		/* wd port offset */
#define	ippi_rdbase	0x203		/* rd port address */
#define	ippi_rdspan	0x200		/* rd port ragne */
#define	ippi_rdskip	0x40		/* rd port scan gap */

static int ippi_map_io __P((struct ippi_softc *, slot_device_res_t, u_int, int));
#if	0
static int ippi_map_mem __P((struct ippi_softc *, slot_device_res_t, u_int, int));
static int ippi_map_dma __P((struct ippi_softc *, slot_device_res_t, u_int, int));
#endif
static int ippi_map_intr __P((struct ippi_softc *, slot_device_res_t, u_int, int));
static void ippi_initres __P((slot_device_res_t));

static struct ippi_softc ippi_softc_tmp;
struct ippi_softc *ippi_gsc;

void
ippiattach(void)
{
	struct ippi_softc *sc = &ippi_softc_tmp;
	long iobase = IPPI_DEFAULT_RDPORT;

	sc->sc_iot = NULL;
	sc->sc_addrh = iobase;
	sc->sc_wdh = iobase + ippi_wdofs;
	ippi_scan_board(sc, 1);
	ippi_gsc = sc;
}

int
ippi_open(product, lun, drp)
	u_long product;
	u_long lun;
	slot_device_res_t *drp;
{
	struct ippi_softc *sc = &ippi_softc_tmp;
	struct ippi_board_info *ib;
	struct ippi_ld_info *il;
	slot_device_res_t dr;
	int i;

	for (i = 0; i < ippi_ndev; i ++)
	{
		ib = sc->sc_ib[i];
		if (ib == NULL)
			continue;
		if (ib->ib_bid.bid_id == product)
			goto found;
	}
	return ENOTTY;

found:
	il = &ib->ib_il[lun];
	if (il->il_status != IPP_FOUND)
		return ENOTTY;

	dr = &il->il_bdr;

	ippi_wakeup_target(sc, ib, il);
	if (ippi_is_active(sc) != 0)
		goto out;

	ippi_activate_target(sc, 1);
	ippi_wfk_target(sc);

	ippi_wakeup_target(sc, ib, il);
	ippi_map_intr(sc, dr, 0, 1);
	for (i = 0; i < SLOT_DEVICE_NIO; i ++)
	{
		ippi_map_io(sc, dr, i, 0);
	}

out:
	ippi_wfk_target(sc);
	*drp = dr;
	return 0;
}

/************************************************************
 * ioctl
 ************************************************************/
static int
ippi_find_resources(sc, cis, ldn, dr)
	struct ippi_softc *sc;
	u_int8_t *cis;
	u_int ldn;
	slot_device_res_t dr;
{
	struct ippres_request ir;
	ippres_res_t irr;
	u_int maxdfn, dfn;
	int pri, error;


	if ((irr = alloc(sizeof(*irr))) == NULL)
		return ENOMEM;

	memset(&ir, 0, sizeof(ir));
	ir.ir_ldn = ldn;
	ippi_initres(&irr->irr_dr);

	error = ippres_parse_items(cis, &ir, irr);
	if (error != 0)
		goto out;

	maxdfn = ir.ir_ndfn;
	for (pri = 0; pri < 3; pri ++)
	{
		for (dfn = 0; dfn == 0 || dfn < maxdfn; dfn ++)
		{
			ippi_initres(&irr->irr_dr);
			ir.ir_dfn = dfn;
			error = ippres_parse_items(cis, &ir, irr);
			if (error != 0)
				goto out;

			if (pri < 3 && irr->irr_pri != pri)
				continue;
		
			error = ippi_space_prefer(sc, irr);
			if (error == 0)
			{
				*dr = irr->irr_dr;
				goto out;
			}
		}
	}

	error = ENOSPC;

out:
	free(irr, sizeof(*irr));
	return error;
}

int
ippi_space_prefer(sc, irr)
	struct ippi_softc *sc;
	ippres_res_t irr;
{
	slot_device_res_t dr = &irr->irr_dr;
	int idx;
	
	for (idx = 0; idx < SLOT_DEVICE_NIM; idx ++)
	{
		if (dr->dr_im[idx].im_hwbase == SLOT_DEVICE_UNKVAL)
			continue;

		if (idx < SLOT_DEVICE_NIO)
		{
			if (dr->dr_im[idx].im_hwbase < 0xd0)
				return EINVAL;	
		}
		else
		{
			/* memory */
			continue;
		}
	}
	return 0;
}

/************************************************************
 * device isolation and kernel registerations
 ************************************************************/
int
ippi_scan_board(sc, init)
	struct ippi_softc *sc;
	int init;
{
	int ndevs;
	u_long gap, eport, sport, port;

	ndevs = 0;
	gap = ippi_rdskip;
	sport = ippi_rdbase;
	eport = sport + ippi_rdspan;

	for (port = sport; port < eport; port += gap)
	{
		sc->rd_port = port / sizeof(u_int32_t);
		sc->sc_rdh = port;
		ndevs = ippi_isolation(sc, init);
		if (ndevs != 0)
		{
			ippi_ndev = ndevs;
			break;
		}
	}

	printf("%s: read port 0x%lx(0x%lx).\n",
		dv_xname, sc->rd_port, port);
	printf("%s: kernel found %d pnp device(s).\n",
		dv_xname, ndevs);
	return 0;
}

#define	IPPI_ISOLATION_BUGGY_WAIT(wait)	delay((wait))

int
ippi_isolation(sc, init)
	struct ippi_softc *sc;
	int init;
{
	bus_space_tag_t iot = sc->sc_iot;
	struct ippi_board_info *ib;
	struct ippi_ld_info *il;
	struct serial_id sdata;
	int i, idx;

	/* Anyway all target should be "Wait for Key" state */
	ippi_wfk_target(sc);
	IPPI_ISOLATION_BUGGY_WAIT(1000);

	/* Wait for key state => sleep state */
	ippi_send_initiation_LFSR(sc);
	IPPI_ISOLATION_BUGGY_WAIT(2000);

	/* Reset CSN of all boards */
	ippi_regw(sc, ippi_ccr, CCR_CSNR);
	IPPI_ISOLATION_BUGGY_WAIT(2000);

	sc->sc_maxb = 0;
	for (idx = 0; idx < MAX_CARDS; idx ++)
	{
#ifdef	IPPI_BUGGY_PROTOCOL_CARDS
		/* once fall back to Wait for Key state */
		ippi_wfk_target(sc);
		IPPI_ISOLATION_BUGGY_WAIT(1000);

		/* Send LFSR and sleep state again */
		ippi_send_initiation_LFSR(sc);
		IPPI_ISOLATION_BUGGY_WAIT(2000);
#endif	/* IPPI_BUGGY_PROTOCOL_CARDS */

		/* Wake[0]: 
	 	 *	sleep state  => isolation state 
		 *	config state => isolation state
		 */
		ippi_regw(sc, ippi_wur, WUR_CONFIG);
		IPPI_ISOLATION_BUGGY_WAIT(1000);

		/* Set RD port */
		ippi_regw(sc, ippi_rdar, sc->rd_port);
		IPPI_ISOLATION_BUGGY_WAIT(1000);

		/* Serialize */
		if (ippi_get_serial(sc, &sdata) == 0)
			break;

		/*
		 * Give a CSN to the serialized board.
		 * isolation state => config state
		 */
		ippi_regw(sc, ippi_csnr, idx + 1);
		bus_space_write_1(iot, sc->sc_addrh, 0, ippi_sr);

		ib = sc->sc_ib[idx];
		if (ib == NULL)
		{
			ib = alloc(sizeof(*ib));
			if (ib == NULL)
				goto out;
			memset(ib, 0, sizeof(*ib));
			sc->sc_ib[idx] = ib;
		}

		ippi_setup_ib(ib, idx + 1, &sdata);
		ippi_read_info(sc, ib, &sdata);

		for (i = 0; i < MAX_LDNS; i ++)
		{
			il = &ib->ib_il[i];
			il->il_ldn = i;
			il->il_flags = 0;
			il->il_status = 0;
			ippi_initres(&il->il_bdr);
		}

		for (i = 0; i < MAX_LDNS; i ++)
		{
			il = &ib->ib_il[i];
			if(ippi_find_resources(sc, ib->ib_cis, i, &il->il_bdr))
			{
				break;
			}

			il->il_status = IPP_FOUND;
			printf("%s: ld(%d:%d) ", dv_xname, ib->ib_csn, i);
			ippi_print(sc, ib);

			ippi_regw(sc, ippi_ldnr, il->il_ldn);
			if (ippi_is_active(sc) != 0)
			{
				il->il_flags |= IPPI_LD_BOOTACTIVE;
				ippi_read_config(sc, &il->il_bdr);
#if	0
				ippi_regw(sc, ippi_irlr, 0);	/* XXX */
#endif	
			}
		}

		ib->ib_maxldn = i;
	}

out:
	ippi_wfk_target(sc);
	sc->sc_maxb = idx;
	return idx;
}

/************************************************************
 * misc utils
 ************************************************************/
void
ippi_initres(dr)
	slot_device_res_t dr;
{
	int i;

	memset(dr, 0, sizeof(*dr));

	dr->dr_ndrq = 2;
	for (i = 0; i < SLOT_DEVICE_NDRQ; i ++)
		dr->dr_drq[i].dc_chan = SLOT_DEVICE_UNKVAL;

	dr->dr_npin = 2;
	for (i = 0; i < SLOT_DEVICE_NPIN; i++)
	{
		dr->dr_pin[i].dc_flags = IRTR_HIGH;
		dr->dr_pin[i].dc_chan = SLOT_DEVICE_UNKVAL;
	}

	dr->dr_nio = SLOT_DEVICE_NIO;
	dr->dr_nmem = SLOT_DEVICE_NMEM;
	dr->dr_nim = SLOT_DEVICE_NIM;
	for (i = 0; i < SLOT_DEVICE_NIM; i++)
	{
		dr->dr_im[i].im_hwbase = SLOT_DEVICE_UNKVAL;
		dr->dr_im[i].im_size = SLOT_DEVICE_UNKVAL;
		if (i >= SLOT_DEVICE_NIO)
		{
			dr->dr_im[i].im_flags = (SDIM_BUS_WIDTH16 | MCR_BUS16);
			dr->dr_im[i].im_type = SLOT_DEVICE_SPMEM;
		}
		else
		{
			dr->dr_im[i].im_flags = 0;
			dr->dr_im[i].im_type = SLOT_DEVICE_SPIO;
		}
	}
}

void
ippi_setup_ib(ib, csn, sbp)
	struct ippi_board_info *ib;
	u_int csn;
	struct serial_id *sbp;
{

	ib->ib_csn = csn;
	ib->ib_bid.bid_id = htonl(sbp->si_product);
	ib->ib_bid.bid_serial = htonl(sbp->si_un);
}

/************************************************************
 * HW primitive access (dirty part)
 ************************************************************/
static void
ippi_regw(sc, addr, val)
	struct ippi_softc *sc;
	bus_addr_t addr;
	u_int8_t val;
{

	bus_space_write_1(sc->sc_iot, sc->sc_addrh, 0, addr);
	bus_space_write_1(sc->sc_iot, sc->sc_wdh,   0, val);
}

static u_int8_t
ippi_regr(sc, addr)
	struct ippi_softc *sc;
	bus_addr_t addr;	
{

	bus_space_write_1(sc->sc_iot, sc->sc_addrh, 0, addr);
	return bus_space_read_1(sc->sc_iot, sc->sc_rdh, 0);
}

static void
ippi_regw_2(sc, offs, val)
	struct ippi_softc *sc;
	bus_addr_t offs;
	u_int16_t val;
{
	
	ippi_regw(sc, offs, (u_int8_t) (val >> NBBY));
	ippi_regw(sc, offs + 1, (u_int8_t) val);
}

static u_int16_t
ippi_regr_2(sc, offs)
	struct ippi_softc *sc;
	bus_addr_t offs;
{
	register u_int16_t val;

	val = ippi_regr(sc, offs);
	val = (val << NBBY) | (u_int16_t) ippi_regr(sc, offs + 1);
	return val;
}

/*
 * Send initiation LFSR as described in "Plug and Play ISA Specification,
 * Intel May 94."
 */
static void
ippi_send_initiation_LFSR(sc)
	struct ippi_softc *sc;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t ioh = sc->sc_addrh;
	u_int key, i;

	for (i = 0; i < 2; i ++)
		bus_space_write_1(iot, ioh, 0, 0);
	for (key = 0x6a, i = 0; i < 32; i++)
	{
		bus_space_write_1(iot, ioh, 0, key);
		key = (key >> 1) | (((key ^ (key >> 1)) << 7) & 0xff);
	}
}

static void
ippi_wakeup_target(sc, ib, il)
	struct ippi_softc *sc;
	struct ippi_board_info *ib;
	struct ippi_ld_info *il;
{

	ippi_send_initiation_LFSR(sc);
	ippi_regw(sc, ippi_wur, ib->ib_csn);
	ippi_regw(sc, ippi_ldnr, il->il_ldn);
}

static void
ippi_wfk_target(sc)
	struct ippi_softc *sc;
{

	ippi_regw(sc, ippi_ccr, CCR_WFK);
}

static void
ippi_activate_target(sc, flags)
	struct ippi_softc *sc;
	int flags;
{

	ippi_regw(sc, ippi_iorcr, 0);
	ippi_regw(sc, ippi_ar, flags ? AR_ACTIVE : 0);
}

static int
ippi_is_active(sc)
	struct ippi_softc *sc;
{

	return (ippi_regr(sc, ippi_ar) & AR_ACTIVE);
}

/*
 * Get the device's serial number. Returns 1 if the serial is valid.
 */
static int
ippi_get_serial(sc, sdata)
	struct ippi_softc *sc;
	struct serial_id *sdata;
{
	bus_space_tag_t iot = sc->sc_iot;
	bus_space_handle_t rdh = sc->sc_rdh;
	u_int i, bit, valid = 0, sum = 0x6a;
	u_int8_t *data;

	bus_space_write_1(iot, sc->sc_addrh, 0, ippi_sir);
	delay(1000);

	data = (u_int8_t *) sdata;
	memset(sdata, 0, sizeof(*sdata));
	for (i = 0; i < NBBY * sizeof(*sdata); i++)
	{
		bit = bus_space_read_1(iot, rdh, 0) == 0x55;
		delay(250);
		bit = (bus_space_read_1(iot, rdh, 0) == 0xaa) && bit;
		delay(250);

		valid = valid || bit;
		if (i < 64)
			sum = (sum >> 1) |
			    (((sum ^ (sum >> 1) ^ bit) << 7) & 0xff);
		data[i / 8] = (data[i / 8] >> 1) | (bit ? 0x80 : 0);
	}

	valid = valid && (sdata->si_cksum == sum);
	return valid;
}

static void
ippi_read_info(sc, ib, sbp)
	struct ippi_softc *sc;
	struct ippi_board_info *ib;
	struct serial_id *sbp;
{
	u_int8_t *ids, *mids;
	int cnt;

 	ids = (u_int8_t *) ib->ib_cis;
	mids = ids + MAX_RDSZ;
	bcopy(sbp, ids, sizeof(*sbp));
	for (ids = ids + sizeof(*sbp); ids < mids; ids ++)
	{
		cnt = 0x100;
		while((ippi_regr(sc, ippi_sr) & SR_READY) == 0)
		{
			if (cnt -- < 0)
				return;
			delay(1);
		}
		*ids = ippi_regr(sc, ippi_rdr);
	}
}

/***************************************************************
 * setup resources or read resources
 ***************************************************************/
#define	IPPI_IOB_MASK (~(0x07U))
#define	IPPI_ADDRESS_LINE 24

int ippi_debug = 0;
void ippi_print_dr __P((slot_device_res_t));

void
ippi_print_dr(dr)
	slot_device_res_t dr;
{
	int idx;

	for (idx = 0; idx < dr->dr_nio; idx ++)
		printf("io[%d]: 0x%lx 0x%lx\n", idx,
			dr->dr_io[idx].im_hwbase,
			dr->dr_io[idx].im_size);

	for (idx = 0; idx < dr->dr_npin; idx ++)
		printf("irq[%d]: 0x%lx 0x%lx\n", idx, 
			dr->dr_pin[idx].dc_chan, 
			dr->dr_pin[idx].dc_flags);

	for (idx = 0; idx < dr->dr_ndrq; idx ++)
		printf("drq[%d]: 0x%lx 0x%lx\n", idx, 
			dr->dr_drq[idx].dc_chan, 
			dr->dr_drq[idx].dc_flags);

	for (idx = 0; idx < dr->dr_nmem; idx ++)
		printf("mem[%d]: 0x%lx 0x%lx 0x%lx\n", idx,
			dr->dr_mem[idx].im_hwbase,
			dr->dr_mem[idx].im_size,
			dr->dr_mem[idx].im_flags);
}

static int
ippi_map_io(sc, dr, idx, unmap)
	struct ippi_softc *sc;
	slot_device_res_t dr;
	u_int idx;
	int unmap;
{
	u_int val;

	if (idx >= MAX_IOWIN)
		return EINVAL;

	if (dr->dr_nio >= SLOT_DEVICE_NIO)
		dr->dr_nio = SLOT_DEVICE_NIO;

	if (idx >= dr->dr_nio || unmap != 0 || 
	    dr->dr_io[idx].im_hwbase == SLOT_DEVICE_UNKVAL)
		val = 0;
	else
		val = (u_int16_t) dr->dr_io[idx].im_hwbase;

	ippi_regw_2(sc, ippi_ioar + IPPI_IOR_SKIP * idx, val);

	return 0;
}

#if	0
static int
ippi_map_mem(sc, dr, idx, unmap)
	struct ippi_softc *sc;
	slot_device_res_t dr;
	u_int idx;
	int unmap;
{

	return 0;
}
#endif

static int
ippi_map_intr(sc, dr, idx, unmap)
	struct ippi_softc *sc;
	slot_device_res_t dr;
	u_int idx;
	int unmap;
{	

	ippi_regw(sc, ippi_irlr, 0);
	return 0;
}

#if	0
static int
ippi_map_dma(sc, dr, idx, unmap)
	struct ippi_softc *sc;
	slot_device_res_t dr;
	u_int idx;
	int unmap;
{	

	ippi_regw(sc, ippi_dcsr, 4);
	return 0;
}
#endif

static int
ippi_read_config(sc, dr)
	struct ippi_softc *sc;
	slot_device_res_t dr;
{
	bus_addr_t offs;
	u_int32_t val;
	int idx;

	/* get io windows resources */
	offs = 0;
	for (idx = 0; idx < SLOT_DEVICE_NIO; idx ++, offs += IPPI_IOR_SKIP)
	{
		if (idx >= MAX_IOWIN)
			continue;

		val = ippi_regr_2(sc, ippi_ioar + offs);
		if (val == 0)
			val = SLOT_DEVICE_UNKVAL;
		dr->dr_io[idx].im_hwbase = val;
	}

	/* get irq resources */
	offs = 0;
	for (idx = 0; idx < SLOT_DEVICE_NPIN; idx ++, offs += IPPI_ICR_SKIP)
	{
		if (idx >= MAX_IRQSET)
			continue;

		val = ippi_regr(sc, ippi_irlr + offs);
		if (val == 0)
			val = SLOT_DEVICE_UNKVAL;
		dr->dr_pin[idx].dc_chan = val;
		dr->dr_pin[idx].dc_flags = ippi_regr(sc, ippi_irtr + offs);
	}

	/* get drq resources */
	offs = 0;
	for (idx = 0; idx < SLOT_DEVICE_NDRQ; idx ++, offs += IPPI_DMAR_SKIP)
	{
		if (idx >= MAX_DRQSET)
			continue;

		val = ippi_regr(sc, ippi_dcsr + offs);
		if (val > 3)
			val = SLOT_DEVICE_UNKVAL;
		dr->dr_drq[idx].dc_chan = val;
	}

	/* get mem resources */
	offs = 0;
	for (idx = 0; idx < SLOT_DEVICE_NMEM; idx ++, offs += IPPI_MEMR_SKIP)
	{
		if (idx >= MAX_MEMWIN)
			continue;

		val = ((u_long) ippi_regr_2(sc, ippi_mar + offs)) << NBBY;
		if (val == 0)
			val = SLOT_DEVICE_UNKVAL;
		dr->dr_mem[idx].im_hwbase = val;
		dr->dr_mem[idx].im_flags = ippi_regr(sc, ippi_mcr + offs);
		val = ((u_long) ippi_regr_2(sc, ippi_mular + offs)) << NBBY;
		if (dr->dr_mem[idx].im_flags & MCR_ULA)
			val -= dr->dr_mem[idx].im_hwbase;
		else if (val != 0)
			val = ((~val) & ((1 << IPPI_ADDRESS_LINE) - 1)) + 1;
		dr->dr_mem[idx].im_size = val;
			
	}

	if (ippi_debug != 0)
		ippi_print_dr(dr);
	return 0;
}

/***************************************************************
 * misc print
 ***************************************************************/
void
ippi_print(sc, ib)
	struct ippi_softc *sc;
	struct ippi_board_info *ib;
{
	u_int8_t *data = ib->ib_cis;

	printf("vendor:%c%c%c%02x%02x(0x%08x) serial: 0x%08x\n",
	    ((data[0] & 0x7c) >> 2) + 64,
	    (((data[0] & 0x03) << 3) | ((data[1] & 0xe0) >> 5)) + 64,
	    (data[1] & 0x1f) + 64, data[2], data[3],
	     ib->ib_bid.bid_id, ib->ib_bid.bid_serial);
}

/***************************************************************
 * parser
 ***************************************************************/
#include <i386/Cbus/pnp/ippres.c>
#endif	/* BOOT_IPPI */
