/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <machine/pio.h>

#include <lib/libsa/stand.h>
#include <libi386.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/ne2000reg.h>
#include "../busio_user.h"
#include "dp8390piosubr.h"
#include "ne.h"

struct ne_switch {
	int (*ne_init) __P((u_char *, void **));
	void (*ne_stop) __P((void *));
};

struct ne_switch *neswp_gc;
void *ne_softc_gc;

static struct ne_switch ne_switch[] = {
	{ ne_isa_init,
	  ne_isa_stop
	},
#ifdef	BOOT_PCCS
	{ ne_pccs_init,
	  ne_pccs_stop
	},
#endif	/* BOOT_PCCS */
#ifdef	BOOT_IPPI
	{ ne_ippi_init,
	  ne_ippi_stop
	},
#endif	/* BOOT_IPPI */
	{ NULL, }
};

int
EtherInit(myadr)
	unsigned char *myadr;
{
	int rv = 0;
	void *m;
	struct ne_switch *nswp = &ne_switch[0];

	for ( ; nswp->ne_init != NULL; nswp ++)
	{
		rv = (*nswp->ne_init) (myadr, &m);
		if (rv != 0)
		{
			neswp_gc = nswp;
			ne_softc_gc = m;
			break;
		}
	}
	return rv;
}

void
EtherStop(void)
{

	if (neswp_gc == NULL)
		return;
	(*neswp_gc->ne_stop) (ne_softc_gc);
}
