/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/* #define	NE_ISA_FAKE_PCCS_DEBUG	1 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <machine/pio.h>

#include <lib/libsa/stand.h>
#include <libi386.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/ne2000reg.h>
#include "../busio_user.h"
#include "dp8390piosubr.h"
#include "../ippivar.h"
#include "ne.h"

#ifdef _STANDALONE
#include <lib/libkern/libkern.h>
#include <bootinfo.h>
#include "../bootinfo_bootflags.h"
#endif

#include "etherdrv.h"

static struct dp8390pio_softc dp8390_softc_tmp;

#ifdef _STANDALONE
static struct btinfo_netif bi_netif;
#endif

static u_long ne_ippi_product[] = {
	0x91839183,
	0x8689f428,
	0x8689f454,
	0x8986f428,
	0
};

int
ne_ippi_init(myadr, argp)
	unsigned char *myadr;
	void **argp;
{
	struct dp8390pio_softc *sc = &dp8390_softc_tmp;
	slot_device_res_t dr;
	int i;

	bzero(sc, sizeof(*sc));
	for (i = 0; ne_ippi_product[i] != 0; i ++)
		if (ippi_open(ne_ippi_product[i], 0, &dr) == 0)
			goto found;
	return 0;

found:
	if (dr->dr_io[0].im_hwbase != SLOT_DEVICE_UNKVAL &&
	    dr->dr_io[1].im_hwbase != SLOT_DEVICE_UNKVAL &&
	    dr->dr_io[0].im_size == 0x10 &&
	    dr->dr_io[1].im_size == 0x10)
	{
		if (dr->dr_io[0].im_hwbase >= dr->dr_io[1].im_hwbase)
		{
			sc->sc_nich = dr->dr_io[1].im_hwbase;
			sc->sc_asich = dr->dr_io[0].im_hwbase;
		}
		else
		{
			sc->sc_nich = dr->dr_io[0].im_hwbase;
			sc->sc_asich = dr->dr_io[1].im_hwbase;
		}
	}
	else
	{
		return 0;
	}

	sc->cr_proto = ED_CR_RD2;
	sc->sc_useword = 1;
	sc->dcr_reg = ED_DCR_FT1 | ED_DCR_LS | ED_DCR_WTS;
	sc->sc_romstart = 0;

	sc->eth_myaddr[0] = 0xff;
	if (dp8390_config(sc, 16384, 16384) != 0)
		return 0;
	bcopy(sc->eth_myaddr, myadr, 6);

#ifdef _STANDALONE
	bi_netif.addr.iobase = dr->dr_io[0].im_hwbase;
	strncpy(bi_netif.ifname, "ne", sizeof(bi_netif.ifname));
	bi_netif.bus = BI_BUS_ISA;
	BI_ADD(&bi_netif, BTINFO_NETIF, sizeof(bi_netif));
#endif
	*argp = sc;
	return 1;
}

void
ne_ippi_stop(arg)
	void *arg;
{
	struct dp8390pio_softc *sc = arg;

	dp8390_stop(sc);
}
