/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define	NE_PCCS_AUTO_DETECT_ADDR 1

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <machine/pio.h>

#include <lib/libsa/stand.h>
#include <libi386.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/ne2000reg.h>
#include "../busio_user.h"
#include "dp8390piosubr.h"
#include "ne.h"
#include "../pccs.h"

#ifdef _STANDALONE
#include <lib/libkern/libkern.h>
#include <bootinfo.h>
#include "../bootinfo_bootflags.h"
#endif

#include "etherdrv.h"

static struct dp8390pio_softc dp8390_softc_tmp;

static int ne_pccs_probe __P((struct card_info *));
#ifdef _STANDALONE
static struct btinfo_netif bi_netif;
#endif

#ifdef	NE_PCCS_AUTO_DETECT_ADDR
struct ne_vendor_tag {
	struct ne_vendor_id_tag *vt_vid;

	int (*vt_get_nd) __P((struct ne_vendor_tag *, \
			      bus_space_tag_t, bus_space_handle_t, u_int8_t *));
	bus_size_t vt_ndofs;
	int vt_nent;
};

struct end_vendor_macaddr {
	u_int8_t mac[3];
};

struct ne_vendor_id_tag {
	u_int vid_vnd;
	u_int vid_vndmask;
};

#define	END_LOOKUP_INVALID	1
#define	END_LOOKUP_NOTFOUND	2
#define	ID_TAG_SIZE(name) (sizeof((name)) / sizeof(struct ne_vendor_id_tag))

static int ne_pccs_nd_from_asic_regs __P((struct ne_vendor_tag *, bus_space_tag_t, bus_space_handle_t, u_int8_t *));
static int ne_pccs_get_nd __P((struct ne_vendor_tag *, bus_space_tag_t, bus_space_handle_t, u_int8_t *));
static int end_lookup __P((u_int8_t *));

static struct ne_vendor_id_tag dl109_vendor_id_tag[] = {
	{	
		0xc00f00ff,
		0xffffff00,
	},
	{	
		0x0149023f,
		0xfffffff0,
	},
	{
		0x0149c1ff,
		0xffffff00,
	},
	{
		0x0105ea15,
		0xffffffff,
	},
};

static struct ne_vendor_id_tag PCCard_vendor_id_tag[] = {
	{
		0xffffffff,
		0x0,
	},
};

struct ne_vendor_tag ne_pisa_vendor_tag[] = {
	{	 &dl109_vendor_id_tag[0],
		ne_pccs_nd_from_asic_regs,
		0x4,
		ID_TAG_SIZE(dl109_vendor_id_tag)
	},

	{	&PCCard_vendor_id_tag[0],
		ne_pccs_get_nd,
		0,
		ID_TAG_SIZE(PCCard_vendor_id_tag)
	},
	
	{	NULL, },
};


static struct end_vendor_macaddr end_valid_macaddr[] = {
	{{ 0x00, 0x00, 0x48 }}, {{ 0x00, 0x00, 0x65 }}, {{ 0x00, 0x00, 0xe8 }},
	{{ 0x00, 0x00, 0xf4 }}, {{ 0x00, 0x04, 0xac }}, {{ 0x00, 0x20, 0xcb }},
	{{ 0x00, 0x20, 0xe0 }}, {{ 0x00, 0x20, 0xe8 }}, {{ 0x00, 0x40, 0x26 }},
	{{ 0x00, 0x40, 0x33 }}, {{ 0x00, 0x40, 0x4c }}, {{ 0x00, 0x40, 0x95 }},
	{{ 0x00, 0x40, 0xc7 }}, {{ 0x00, 0x40, 0xf6 }}, {{ 0x00, 0x47, 0x43 }},
	{{ 0x00, 0x80, 0x19 }}, {{ 0x00, 0x80, 0x45 }}, {{ 0x00, 0x80, 0x4c }},
	{{ 0x00, 0x80, 0xc6 }}, {{ 0x00, 0x80, 0x98 }},
	{{ 0x00, 0x80, 0xc8 }}, {{ 0x00, 0x90, 0xfe }}, {{ 0x00, 0xa0, 0x0c }},
	{{ 0x00, 0xa0, 0xb0 }}, {{ 0x00, 0xc0, 0xf0 }}, {{ 0x08, 0x00, 0x00 }},
	{{ 0x08, 0x00, 0x17 }}, {{ 0x08, 0x00, 0x42 }}, {{ 0x08, 0x00, 0x5a }},
	{{ 0x00, 0x00, 0x4c }}, {{ 0x00, 0x80, 0x43 }},
	{{ 0x00, 0xe0, 0x98 }}, {{ 0x00, 0x40, 0x41 }}, {{ 0x00, 0x40, 0xb4 }},
	{{ 0xff, 0xff, 0xff }}	/* terminate */
};

struct ne_nicoffset {
	u_int offset;
};

static struct ne_nicoffset ne_valid_nicoffs[] = {
	{ 0x07f8 }, { 0x0088 }, { 0x00e0 }, { 0x01ba },
	{ 0x00e4 }, { 0x001d }, { 0x0020 }, { 0x03f8 },
	{ 0 }	/* terminate */
};

static int
end_lookup(ndp)
	u_int8_t *ndp;
{
	struct end_vendor_macaddr *macp;

	if (ndp[0] == 0xff || (ndp[0] | ndp[1] | ndp[2]) == 0 || ndp[1] & 1)
		return END_LOOKUP_INVALID;

	macp = &end_valid_macaddr[0];
	for ( ; macp->mac[0] != 0xff; macp ++)
		if (bcmp(ndp, macp->mac, 3) == 0)
			return 0;

	return END_LOOKUP_NOTFOUND;
}


static int
ne_pccs_get_nd(vtp, asict, asich, enaddrp)
	struct ne_vendor_tag *vtp;
	bus_space_tag_t asict;
	bus_space_handle_t asich;
	u_int8_t *enaddrp;
{
	u_int8_t *p;
	struct ne_nicoffset *np;

	p = cisbufp;
	if (p == NULL)
		return ENOENT;

	for (np = &ne_valid_nicoffs[0]; np->offset != 0; np ++)
	{
		if (end_lookup(&p[np->offset]) == 0)
		{
			p = &p[np->offset];
			bcopy(p, enaddrp, 6);
			return 0;
		}
	}
	return ENOENT;
}

static int
ne_pccs_nd_from_asic_regs(vtp, asict, asich, p)
	struct ne_vendor_tag *vtp;
	bus_space_tag_t asict;
	bus_space_handle_t asich;
	u_int8_t *p;
{
	bus_size_t ofs = vtp->vt_ndofs;
	int i;

	for (i = 0; i < 6; i ++)
		p[i] = bus_space_read_1(asict, asich, ofs + i);
	return end_lookup(p);
}

static struct ne_vendor_tag *
ne_pisa_find_vt(ci)
	struct card_info *ci;
{
	struct ne_vendor_tag *vtp;
	struct ne_vendor_id_tag *vidp;
	u_int id, i;

	for (vtp = &ne_pisa_vendor_tag[0]; vtp->vt_vid != NULL; vtp ++)
	{
		vidp = vtp->vt_vid;
		for (i = 0; i < vtp->vt_nent; i++, vidp ++)
		{
			id = ci->ci_manfid | (~vidp->vid_vndmask);
			if (id == vidp->vid_vnd)
				return vtp;
		}
	}
	return NULL;
}
#endif	/* NE_PCCS_AUTO_DETECT_ADDR */

static int
ne_pccs_probe(ci)
	struct card_info *ci;
{

	return 1;
}

int
ne_pccs_init(myadr, argp)
	unsigned char *myadr;
	void **argp;
{
	static struct card_info citmp;
	struct dp8390pio_softc *sc = &dp8390_softc_tmp;
	struct card_info *ci = &citmp;
	struct bootflags_device *bdp;
	slot_handle_t sid = 0;
	int i;
#ifdef	NE_PCCS_AUTO_DETECT_ADDR
	static struct ne_vendor_tag *vtp;
#endif	/* NE_PCCS_AUTO_DETECT_ADDR */

	bzero(ci, sizeof(*ci));
	bzero(sc, sizeof(*sc));
	if (pccs_open(sid, ci) != 0)
		return 0;

	if (ne_pccs_probe(ci) == 0)
		goto bad;

	if (pccs_card_open(sid, ci) != 0)
		goto bad;

	if (ci->ci_res.dr_io[0].im_hwbase != SLOT_DEVICE_UNKVAL &&
	    ci->ci_res.dr_io[1].im_hwbase == SLOT_DEVICE_UNKVAL &&
	    ci->ci_res.dr_io[0].im_size == 0x20)
	{
		sc->sc_nich = ci->ci_res.dr_io[0].im_hwbase;
		sc->sc_asich = ci->ci_res.dr_io[0].im_hwbase + 0x10;
	}
	else if (ci->ci_res.dr_io[0].im_hwbase != SLOT_DEVICE_UNKVAL &&
	         ci->ci_res.dr_io[1].im_hwbase != SLOT_DEVICE_UNKVAL &&
		 ci->ci_res.dr_io[0].im_size == 0x10 &&
		 ci->ci_res.dr_io[1].im_size == 0x10)
	{
		sc->sc_nich = ci->ci_res.dr_io[0].im_hwbase;
		sc->sc_asich = ci->ci_res.dr_io[1].im_hwbase;
	}
	else
	{
		goto bad;
	}

	sc->cr_proto = ED_CR_RD2;
	sc->sc_useword = 1;
	sc->dcr_reg = ED_DCR_FT1 | ED_DCR_LS | ED_DCR_WTS;
	sc->sc_romstart = 0;

#ifdef	NE_PCCS_AUTO_DETECT_ADDR
	vtp = ne_pisa_find_vt(ci);
	if (vtp == NULL)
		goto bad;
	if (vtp->vt_get_nd(vtp, sc->sc_asict, sc->sc_asich, sc->eth_myaddr) != 0)
	{
		sc->eth_myaddr[0] = 0xff;
	}
#else	/* NE_PCCS_AUTO_DETECT_ADDR */
	sc->eth_myaddr[0] = 0xff;
#endif	/* NE_PCCS_AUTO_DETECT_ADDR */

	bdp = lookup_bootflags_device(BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT |
				      BOOTFLAGS_DEVICE_TYPE_BUS_ISA,
				      "ne.eth.addr");
	if (bdp != NULL)
	{
		for (i = 0; i < 6; i ++)
			sc->eth_myaddr[i] = (u_int8_t) bdp->bd_loc[i];
	}

	if (dp8390_config(sc, 16384, 16384) != 0)
		return 0;
	bcopy(sc->eth_myaddr, myadr, 6);

#ifdef _STANDALONE
	bi_netif.addr.iobase = ci->ci_res.dr_io[0].im_hwbase;
	strncpy(bi_netif.ifname, "ne", sizeof(bi_netif.ifname));
	bi_netif.bus = BI_BUS_ISA;
	BI_ADD(&bi_netif, BTINFO_NETIF, sizeof(bi_netif));
#endif

	*argp = sc;
	return 1;

bad:
	pccs_close(sid, ci);
	return 0;
}

void
ne_pccs_stop(arg)
	void *arg;
{
	struct dp8390pio_softc *sc = arg;

	dp8390_stop(sc);
}

