/*	$NetBSD$	*/
#ifdef	BOOT_PCCS_NSP

/*
 *  Copyright (c) 1998, 1999, 2000, 2001
 *	NetBSD/pc98 porting staff. All rights reserved.
 *
 *  Copyright (c) 1998, 1999, 2000, 2001
 *	Naofumi HONDA. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <sys/disk.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <lib/libsa/stand.h>
#include <lib/libkern/libkern.h>
#include <libi386.h>

#include "busio_user.h"
#include "pccs.h"
#include "disk_direct.h"
#include "nsp_pccs.h"

#include <i386/Cbus/dev/nspreg.h>

#define	NSP_SELTIMEOUT		(200 * 1000)
#define	NSP_DELAY_MAX		(2 * 1000 * 1000)
#define	NSP_DELAY_INTERVAL	(1)

struct slccb {
	u_int cb_id;
	u_int cb_lun;

	u_int8_t cb_cmd[16];
	int cb_cmdlen;
	u_int8_t *cb_data;
	int cb_datalen;

	int cb_done;
	u_int8_t cb_status;
};

struct nsp_softc {
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;

	u_int sc_cnt;

	u_int sc_target;/* XXX */
	u_int sc_lun;	/* XXX */
};
/**************************************************************
 * DECLARE
 **************************************************************/
static void nsp_pio_read __P((struct nsp_softc *, struct slccb *));
static void nsphw_init __P((struct nsp_softc *));
static int nsphw_start_selection __P((struct nsp_softc *sc, struct slccb *));
static u_int nsp_fifo_count __P((struct nsp_softc *));
static int nsp_negate_signal __P((struct nsp_softc *, u_int8_t, u_char *));
static int nsp_read_fifo __P((struct nsp_softc *, struct slccb *));
static int nsp_phase_match __P((struct nsp_softc *, u_int8_t, u_int8_t));
static void nsp_delay __P((int));
static int nspintr __P((struct nsp_softc *, struct slccb *));

/****************************************************
 * hwfuncs
 ****************************************************/
static __inline u_int8_t nsp_cr_read_1 __P((bus_space_tag_t bst, bus_space_handle_t bsh, bus_addr_t ofs));
static __inline void nsp_cr_write_1 __P((bus_space_tag_t bst, bus_space_handle_t bsh, bus_addr_t ofs, u_int8_t va));

static __inline u_int8_t
nsp_cr_read_1(bst, bsh, ofs)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_addr_t ofs;
{
	
	bus_space_write_1(bst, bsh, nsp_idxr, ofs);
	return bus_space_read_1(bst, bsh, nsp_datar);
}

static __inline void 
nsp_cr_write_1(bst, bsh, ofs, va)
	bus_space_tag_t bst;
	bus_space_handle_t bsh;
	bus_addr_t ofs;
	u_int8_t va;
{

	bus_space_write_1(bst, bsh, nsp_idxr, ofs);
	bus_space_write_1(bst, bsh, nsp_datar, va);
}
	
static void
nsp_delay(count)
	int count;
{

	count += 2;
	while (count -- > 0)
		outb(0x5f, 0);
}

static void
nsphw_init(sc)
	struct nsp_softc *sc;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	u_int8_t icr;
	static int nsp_bus_reset;

	icr = (SCIENR_SCCHG | SCIENR_RESEL | SCIENR_RST);

	/* block all interrupts */
	bus_space_write_1(bst, bsh, nsp_irqcr, IRQCR_ALLMASK);

	/* setup SCSI interface */
	bus_space_write_1(bst, bsh, nsp_ifselr, IFSELR_IFSEL);

	nsp_cr_write_1(bst, bsh, NSPR_SCIENR, 0);

	nsp_cr_write_1(bst, bsh, NSPR_XFERMR, XFERMR_IO8);
	nsp_cr_write_1(bst, bsh, NSPR_CLKDIVR, CLKDIVR_20M);

	nsp_cr_write_1(bst, bsh, NSPR_SCIENR, icr);
	nsp_cr_write_1(bst, bsh, NSPR_PARITYR, 0);
	nsp_cr_write_1(bst, bsh, NSPR_PTCLRR,
		       PTCLRR_ACK | PTCLRR_REQ | PTCLRR_HOST | PTCLRR_RSS);

	/* setup fifo asic */
	bus_space_write_1(bst, bsh, nsp_ifselr, IFSELR_REGSEL);
	nsp_cr_write_1(bst, bsh, NSPR_TERMPWRC, 0);
	if ((nsp_cr_read_1(bst, bsh, NSPR_OCR) & OCR_TERMPWRS) == 0)
		nsp_cr_write_1(bst, bsh, NSPR_TERMPWRC, TERMPWRC_POWON);

	nsp_cr_write_1(bst, bsh, NSPR_XFERMR, XFERMR_IO8);
	nsp_cr_write_1(bst, bsh, NSPR_CLKDIVR, CLKDIVR_40M);
	nsp_cr_write_1(bst, bsh, NSPR_TIMERCNT, 0);
	nsp_cr_write_1(bst, bsh, NSPR_TIMERCNT, 0);

	nsp_cr_write_1(bst, bsh, NSPR_SYNCR, 0);
	nsp_cr_write_1(bst, bsh, NSPR_ACKWIDTH, 0);

	/* enable interrupts and ack them */
	nsp_cr_write_1(bst, bsh, NSPR_SCIENR, icr);
	bus_space_write_1(bst, bsh, nsp_irqcr, IRQSR_MASK);
	nsp_delay(1000);	/* 1ms */
	if (nsp_bus_reset == 0)
	{
		int i;

		nsp_bus_reset = 1;
		printf("nsp: bus reset ....\n");
		bus_space_write_1(bst, bsh, nsp_irqcr, IRQCR_ALLMASK);

		nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, SCBUSCR_RST);
		nsp_delay(100 * 1000);	/* 100ms */
		nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, 0);
		for (i = 0; i < 5; i ++)
			(void) nsp_cr_read_1(bst, bsh, NSPR_IRQPHS);

		bus_space_write_1(bst, bsh, nsp_irqcr, IRQSR_MASK);
		nsp_delay(1000 * 1000);	/* 1sec */
	}
}

static int
nsphw_start_selection(sc, cb)
	struct nsp_softc *sc;
	struct slccb *cb;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	int wc;
	register u_int8_t arbs, ph;

	/* wait for a bus free */
	sc->sc_cnt = 0;
	while (1)
	{
		ph = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
		if (ph == SCBUSMON_FREE)
			break;
	}

	nsp_cr_write_1(bst, bsh, NSPR_PTCLRR,
		PTCLRR_PT | PTCLRR_ACK | PTCLRR_REQ | PTCLRR_HOST);

	/* start arbitration */
	nsp_cr_write_1(bst, bsh, NSPR_ARBITS, ARBITS_EXEC);

	do 
	{
		arbs = nsp_cr_read_1(bst, bsh, NSPR_ARBITS);
		nsp_delay(1);
	} 
	while ((arbs & (ARBITS_WIN | ARBITS_FAIL)) == 0 && wc -- > 0);

	if ((arbs & ARBITS_WIN) == 0)
	{
		nsp_cr_write_1(bst, bsh, NSPR_ARBITS, ARBITS_CLR);
		return ENODEV;
	}

	nsp_delay(3);
	nsp_cr_write_1(bst, bsh, NSPR_DATA, (1 << 7) | (1 << cb->cb_id));
	nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR,
			SCBUSCR_SEL | SCBUSCR_BSY | SCBUSCR_ATN);
	nsp_delay(3);
	nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, SCBUSCR_SEL |
			SCBUSCR_BSY | SCBUSCR_DOUT | SCBUSCR_ATN);
	nsp_cr_write_1(bst, bsh, NSPR_ARBITS, ARBITS_CLR);
	nsp_delay(3);
	nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR,
		       SCBUSCR_SEL | SCBUSCR_DOUT | SCBUSCR_ATN);
	nsp_delay(1);

	for (wc = 0; wc < NSP_SELTIMEOUT / NSP_DELAY_INTERVAL; wc ++)
	{
		ph = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
		if ((ph & SCBUSMON_BSY) == 0)
		{
			nsp_delay(NSP_DELAY_INTERVAL);
			continue;
		}

		nsp_delay(1);
		ph = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
		if ((ph & SCBUSMON_BSY) != 0)
		{
			nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, SCBUSCR_ATN);
			nsp_delay(1);
			nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, 
				SCBUSCR_ATN | SCBUSCR_ADIR | SCBUSCR_ACKEN);
			nsp_delay(1);
			nsp_cr_write_1(bst, bsh, NSPR_XFERMR,
				       XFERMR_XEN | XFERMR_IO8);
			return 0;
		}
	}

	nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR, 0);
	return ENODEV;
}

static u_int
nsp_fifo_count(sc)
	struct nsp_softc *sc;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	u_int count;

	nsp_cr_write_1(bst, bsh, NSPR_PTCLRR, PTCLRR_RSS_ACK | PTCLRR_PT);
	count = bus_space_read_1(bst, bsh, nsp_datar);
	count += (((u_int) bus_space_read_1(bst, bsh, nsp_datar)) << 8);
	count += (((u_int) bus_space_read_1(bst, bsh, nsp_datar)) << 16);
	return count;
}

static int
nsp_read_fifo(sc, cb)
	struct nsp_softc *sc;
	struct slccb *cb;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	u_int res, len;

	res = nsp_fifo_count(sc);
	if (res == sc->sc_cnt)
		return 0;

	res = res - sc->sc_cnt;
	if (res > cb->cb_datalen)
		res = cb->cb_datalen;

	for (len = 0; len < res; len ++)
		cb->cb_data[len] = bus_space_read_1(bst, bsh, nsp_fifodr);

	cb->cb_data += res;
	cb->cb_datalen -= res;
	sc->sc_cnt += res;
	return 0;
}

static void
nsp_pio_read(sc, cb)
	struct nsp_softc *sc;
	struct slccb *cb;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	register u_int8_t stat, fstat;

	while (1)
	{
		stat = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
		if (stat == (u_int8_t) -1)
			return;

		if ((stat & SCBUSMON_PHMASK) != PHASE_DATAIN)
		{
			nsp_read_fifo(sc, cb);
			return;
		}

		/* data phase */
		fstat = bus_space_read_1(bst, bsh, nsp_fifosr);
		if ((fstat & FIFOSR_FULLEMP) != 0)
		{
			nsp_read_fifo(sc, cb);
		}
		else
		{
			if (cb->cb_datalen <= 0)
				return;
		}
		nsp_delay(1);
	}
}

static int
nsp_negate_signal(sc, mask, s)
	struct nsp_softc *sc;
	u_int8_t mask;
	u_char *s;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	int wc;
	u_int8_t regv;

	for (wc = 0; wc < NSP_DELAY_MAX / NSP_DELAY_INTERVAL; wc ++)
	{
		regv = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
		if (regv == (u_int8_t) -1)
			return -1;
		if ((regv & mask) == 0)
			return 1;
		nsp_delay(NSP_DELAY_INTERVAL);
	}
	return -1;
}

/**************************************************************
 * SEQUENCER
 **************************************************************/
static void nsp_error __P((struct nsp_softc *, u_char *, u_int8_t, u_int8_t, u_int8_t));

static void
nsp_error(sc, s, isrc, ph, irqphs)
	struct nsp_softc *sc;
	u_char *s;
	u_int8_t isrc, ph, irqphs;
{

	printf("nsp: %s\n", s);
	printf("nsp: isrc 0x%x scmon 0x%x irqphs 0x%x\n",
	       (u_int) isrc, (u_int) ph, (u_int) irqphs);
}

static int
nsp_phase_match(sc, phase, stat)
	struct nsp_softc *sc;
	u_int8_t phase;
	u_int8_t stat;
{

	if ((stat & SCBUSMON_PHMASK) != phase)
	{
		printf("nsp: phase mismatch 0x%x != 0x%x\n",
			(u_int) phase, (u_int) stat);
		return EINVAL;
	}

	if ((stat & SCBUSMON_REQ) == 0)
		return EINVAL;

	return 0;
}

static int
nspintr(sc, cb)
	struct nsp_softc *sc;
	struct slccb *cb;
{
	bus_space_tag_t bst = sc->sc_iot;
	bus_space_handle_t bsh = sc->sc_ioh;
	int len;
	u_int8_t isrc, ph, irqphs, regv;

	/*******************************************
	 * interrupt check
	 *******************************************/
	bus_space_write_1(bst, bsh, nsp_irqcr, IRQCR_IRQDIS);
	isrc = bus_space_read_1(bst, bsh, nsp_irqsr);
	if (isrc == (u_int8_t) -1 || (isrc & IRQSR_SCSI) == 0)
	{
		bus_space_write_1(bst, bsh, nsp_irqcr, 0);
		return 0;
	}

	ph = nsp_cr_read_1(bst, bsh, NSPR_SCBUSMON);
	irqphs = nsp_cr_read_1(bst, bsh, NSPR_IRQPHS);

	regv = 0;
	if ((isrc & IRQSR_TIMER) != 0)
	{
		regv |= IRQCR_TIMERCL;
	}

	/* check a fifo interrupt */
	if ((isrc & IRQSR_FIFO) != 0)
	{
		regv |= IRQCR_FIFOCL;
	}

	/* clear interrupts */
	bus_space_write_1(bst, bsh, nsp_irqcr, regv);

#ifdef	NSP_DEBUG
	nsp_error(sc, "check", isrc, ph, irqphs);
#endif	/* NSP_DEBUG */

	/*******************************************
	 * Parse hardware SCSI irq reasons register
	 *******************************************/
	if ((irqphs & IRQPHS_RST) != 0)
		return 1;

	if ((irqphs & IRQPHS_RSEL) != 0)
		return 1;

	if ((irqphs & (IRQPHS_PCHG | IRQPHS_LBF)) == 0)
		return 1; 

	/*******************************************
	 * scsi seq
	 *******************************************/
	if ((irqphs & IRQPHS_LBF) != 0 || ph == 0)
	{
		nsp_cr_write_1(sc->sc_iot, sc->sc_ioh, NSPR_XFERMR, XFERMR_IO8);
		cb->cb_done = 1;
		return 1;
	}

	/* check normal scsi phase */
	switch (irqphs & IRQPHS_PHMASK)
	{
	case IRQPHS_CMD:
		if (nsp_phase_match(sc, PHASE_CMD, ph) != 0)
			return 1;

		nsp_cr_write_1(bst, bsh, NSPR_CMDCR, CMDCR_PTCLR);
		for (len = 0; len < cb->cb_cmdlen; len ++)
			nsp_cr_write_1(bst, bsh, NSPR_CMDDR,
				       cb->cb_cmd[len]);

		nsp_cr_write_1(bst, bsh, NSPR_CMDCR, CMDCR_PTCLR | CMDCR_EXEC);
		break;

	case IRQPHS_DATAOUT:
		break;

	case IRQPHS_DATAIN:
		nsp_pio_read(sc, cb);
		break;

	case IRQPHS_STATUS:
		if (nsp_phase_match(sc, PHASE_STATUS, ph) != 0)
			return 1;

		cb->cb_status = nsp_cr_read_1(bst, bsh, NSPR_DATAACK);
		nsp_negate_signal(sc, SCBUSMON_ACK, "xfer<ACK>");
		break;

	case IRQPHS_MSGOUT:
		if (nsp_phase_match(sc, PHASE_MSGOUT, ph) != 0)
			return 1;

		regv = 0x80 | cb->cb_lun;
		nsp_cr_write_1(bst, bsh, NSPR_SCBUSCR,
			       SCBUSCR_ADIR | SCBUSCR_ACKEN);
		nsp_cr_write_1(bst, bsh, NSPR_DATAACK, regv);
		nsp_negate_signal(sc, SCBUSMON_ACK, "xfer<ACK>");
		break;

	case IRQPHS_MSGIN:
		if (nsp_phase_match(sc, PHASE_MSGIN, ph) != 0)
			return 1;

		(void) nsp_cr_read_1(bst, bsh, NSPR_DATAACK);
		nsp_negate_signal(sc, SCBUSMON_ACK, "xfer<ACK>");
		break;

	default:
		nsp_error(sc, "unknown scsi phase", isrc, ph, irqphs);
		break;
	}

	return 1;
}

/****************************************************
 * pccs interface
 ****************************************************/
int
nsp_pccs_read(arg, blkno, nblks, bufp)
	void *arg;
	daddr_t blkno;
	daddr_t nblks;
	u_int8_t *bufp;
{
	struct nsp_softc *sc = arg;
	struct slccb slccb;
	u_int target, lun;
	int i;

	target = sc->sc_target;
	lun = sc->sc_lun;
	bzero(&slccb, sizeof(slccb));
	slccb.cb_id = target;
	slccb.cb_lun = lun;
	slccb.cb_cmd[0] = 0x28;
	slccb.cb_cmd[5] = blkno & 0xff;
	slccb.cb_cmd[4] = (blkno >> 8) & 0xff;
	slccb.cb_cmd[3] = (blkno >> 16) & 0xff;
	slccb.cb_cmd[2] = (blkno >> 24) & 0xff;
	slccb.cb_cmd[8] = nblks & 0xff;
	slccb.cb_cmd[7] = (nblks >> 8) & 0xff;
	slccb.cb_cmdlen = 10;

	for (i = 0; i < 4; i ++)
	{
		slccb.cb_datalen = nblks * 512;	/* XXX */
		slccb.cb_data = bufp;
		if (nsphw_start_selection(sc, &slccb) != 0)
			return EIO;

		slccb.cb_done = 0;
		slccb.cb_status = 0xff;
		while (slccb.cb_done == 0)
		{
			nspintr(sc, &slccb);
			nsp_delay(1000);
		}

		if (slccb.cb_status == 0)
			return 0;
	}

	return EIO;
}

static struct nsp_softc nsptab;

void *
nsp_pccs_attach(da)
	struct disk_attach_args *da;
{
	struct nsp_softc *sc = &nsptab;
	static struct card_info citmp;
	struct card_info *ci;
	slot_handle_t sid = 0; 	/* XXX */
	int drive;
	u_int8_t nsd;

	drive = (da->da_dev & 0x7);
	nsd = ((u_int8_t) get_biosparam(0x482));
	if ((nsd & (1 << drive)) != 0)
		return NULL;

 	ci = &citmp;
	bzero(ci, sizeof(*ci));
	bzero(sc, sizeof(*sc));

	if (pccs_open(sid, ci) != 0)
		return NULL;

	if (strncmp(ci->ci_product, "WBT NinjaSCSI-3", 15) != 0 &&
	    strncmp(ci->ci_product, "IO DATA CBSC16", 14) != 0)
		goto bad;

	if (pccs_card_open(sid, ci) != 0)
		goto bad;

	sc->sc_ioh = ci->ci_res.dr_io[0].im_hwbase;
	sc->sc_target = drive;
	sc->sc_lun = 0;			/* XXX */
	nsphw_init(sc);

	da->da_cyl = 0xffff;			/* XXX */
	da->da_head = 8;
	da->da_sec = 32;
	return sc;

bad:	
	pccs_close(sid, ci);
	return NULL;
}
#endif	/* BOOT_PCCS_NSP */
