/*	$NetBSD$	*/
#ifdef	BOOT_ISA_WDC
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * DMA and multi-sector PIO handling are derived from code contributed by
 * Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles M. Hannum.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <sys/disk.h>
#include <sys/syslog.h>
#include <sys/errno.h>

#include <lib/libsa/stand.h>
#include <lib/libkern/libkern.h>
#include <libi386.h>

#include "busio_user.h"
#include "disk_direct.h"
#include "wdcsubr.h"
#include "wdc_isa.h"

#include "../lib/bootinfo.h"
#include "../lib/bootinfo_bootflags.h"

static bus_size_t iat_base2[] = {0, 2, 4, 6, 8, 10, 12, 14};
static void wdc_select_bank_NEC __P((struct channel_softc *, u_int));

static struct wdc_hw wdc_nec98_hardware = {
	"NEC98",
	WDC_DTYPE_GENERIC,

	WDC_ACCESS_PLAIN,
	2,

	{ NULL,
	  NULL,
	  wdc_select_bank_NEC,
	  wdc_cr0_read_plain,
	  wdc_cr0_write_plain,
	  wdc_cr1_read_plain,
	  wdc_cr1_write_plain,
	  wdc_data_read_generic,
	  wdc_data_write_generic
	},

	/* WDC_CAPABILITY_DATA16 | WDC_CAPABILITY_MODE */ 0,
	3,
	0
};


static void
wdc_select_bank_NEC(chp, bank)
	struct channel_softc *chp;
	u_int bank;
{

	if (bank == WDC_BANK_UNKNOWN)
		return;

	bus_space_write_1(chp->ch_bkbst, chp->ch_bkbsh, 0,
			  WDC_BANK_NO(chp, bank));
}

#define	WDC_NEC_IOBASE	0x640
#define	WDC_NEC_BANKCTL	0x432

void *
wdc_isa_attach(da)
	struct disk_attach_args *da;
{
	static struct channel_softc chtmp;
	static struct wd_softc wdtmp;
	struct channel_softc *chp;
	struct wd_softc *wd;
	struct disklabel *lp;
	int drive = da->da_drive;
	u_long biosflags;

	biosflags = get_biosparam(BOOTFLAGS_BIOS_OFFSET);
	if ((biosflags & BOOTFLAGS_FLAGS_DISK_DIRECT) == 0)
		return NULL;

	if (drive >= ATABUS_MAXDRIVE)
		return NULL;

	wd = &wdtmp;
	chp = &chtmp;
	bzero(wd, sizeof(*wd));
	bzero(chp, sizeof(*chp));

	chp->channel = drive;
	chp->cmd_ioh = 0x640;
	chp->ctl_ioh = 0x640 + 0x206;

	if (chp->cmd_ioh == WDC_NEC_IOBASE)
	{
		chp->ch_bkbsh = WDC_NEC_BANKCTL;
		chp->ch_bank = 1;
	}
	chp->ctl_iot = chp->cmd_iot= iat_base2;
	chp->ch_bkbst = NULL;
	if (wdcprobe(chp, &wdc_nec98_hardware) == 0)
		goto bad;

	wd->sc_chp = chp;
	if (whdattach(wd) != 0)
		goto bad;

 	lp = wd->sc_dk.dk_label;
	da->da_cyl = lp->d_ncylinders;
	da->da_head = lp->d_ntracks;
	da->da_sec = lp->d_nsectors;
	return wd;

bad:	
	return NULL;
}

int
wdc_isa_read(arg, blkno, nblks, bufp)
	void *arg;
	daddr_t blkno;
	daddr_t nblks;
	u_int8_t *bufp;
{
	struct wd_softc *wd = arg;
	struct disklabel *lp = wd->sc_dk.dk_label;
	daddr_t i;
	int error = 0;

	for (i = 0; i < nblks && error == 0; i ++)
	{
		error = whdread(wd, blkno, bufp);
		bufp += lp->d_secsize;
		blkno ++;
	}

	return error;
}
#endif	/* BOOT_ISA_WDC */
