/*	$NetBSD$	*/
#ifdef	BOOT_PCCS_WDC
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * DMA and multi-sector PIO handling are derived from code contributed by
 * Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles M. Hannum.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <sys/disk.h>
#include <sys/syslog.h>
#include <sys/errno.h>

#include <lib/libsa/stand.h>
#include <lib/libkern/libkern.h>
#include <libi386.h>

#include "pccs.h"
#include "disk_direct.h"
#include "wdcsubr.h"
#include "wdc_pccs.h"

#include "../lib/bootinfo.h"
#include "../lib/bootinfo_bootflags.h"


#define	WDC_PCCS_DEBUG(mesg)	{printf(mesg);}

static void *__wdc_pccs_attach __P((slot_handle_t, struct disk_attach_args *));

int
wdc_pccs_primary(biosdev)
	int biosdev;
{
	struct disk_attach_args dab;
	u_int8_t nwd;
	int i, nhd;

	bzero(&dab, sizeof(dab));
	dab.da_dev = DISK_DIRECT_DEV_UNK;
	if (wdc_pccs_attach(&dab) == NULL)
		return biosdev;

	biosdev = 0x80;
	nwd = ((u_int8_t) get_biosparam(0x55d)) & 0xf;
	for (nhd = i = 0; i < 4; i++, nwd = nwd >> 1)
	{
		if (nwd & 1)
		{
			nhd++;
		}
	}

	return (biosdev + nhd);
}

void *
wdc_pccs_attach(da)
	struct disk_attach_args *da;
{
	slot_handle_t sid;
	int lun;
	void *wd;
	u_int8_t nwd;

	if (da->da_dev != DISK_DIRECT_DEV_UNK)
	{
		lun = (da->da_dev & 0x3);
		nwd = ((u_int8_t) get_biosparam(0x55d)) & 0xf;
		if ((nwd & (1 << lun)) != 0)
			return NULL;
	}

	for (sid = 0; sid < PCCS_MAX_TARGET_SLOT; sid ++)
	{
		if ((wd = __wdc_pccs_attach(sid, da)) != NULL)
			return wd;
	}
	return NULL;
}

static void *
__wdc_pccs_attach(sid, da)
	slot_handle_t sid;
	struct disk_attach_args *da;
{
	static struct card_info citmp;
	static struct channel_softc chtmp;
	static struct wd_softc wdtmp;
	struct card_info *ci;
	struct channel_softc *chp;
	struct wd_softc *wd;
	struct disklabel *lp;
	int drive = da->da_drive;

	if (drive >= ATABUS_MAXDRIVE)
		return NULL;

	/* XXX 
	 * Always drive no to 0
	 */
	drive = 0;

 	ci = &citmp;
	wd = &wdtmp;
	chp = &chtmp;
	bzero(ci, sizeof(*ci));
	bzero(wd, sizeof(*wd));
	bzero(chp, sizeof(*chp));
	chp->channel = drive;

	if (pccs_open(sid, ci) != 0)
		return NULL;

	if (ci->ci_function != DV_FUNCID_FIXED)
	{
		if  (!(ci->ci_function == 0 && 
		      (strncmp(ci->ci_product, "WIT IDE16", 9) == 0 ||
		       strncmp(ci->ci_product, "IO DATA PCIDE", 13) == 0)))
			goto bad;
	}
	WDC_PCCS_DEBUG("wdc_pccs: function IDE found\n")

	if (pccs_card_open(sid, ci) != 0)
		goto bad;

	if (ci->ci_res.dr_io[0].im_hwbase != SLOT_DEVICE_UNKVAL &&
	    ci->ci_res.dr_io[1].im_hwbase == SLOT_DEVICE_UNKVAL &&
	    ci->ci_res.dr_io[0].im_size == 0x10)
	{
		chp->cmd_ioh = ci->ci_res.dr_io[0].im_hwbase;
		chp->ctl_ioh = ci->ci_res.dr_io[0].im_hwbase + 0x08;
	}
	else if (ci->ci_res.dr_io[0].im_hwbase != SLOT_DEVICE_UNKVAL &&
	         ci->ci_res.dr_io[1].im_hwbase != SLOT_DEVICE_UNKVAL &&
		 ci->ci_res.dr_io[0].im_size == 0x8)
	{
		chp->cmd_ioh = ci->ci_res.dr_io[0].im_hwbase;
		chp->ctl_ioh = ci->ci_res.dr_io[1].im_hwbase;
	}
	else
	{
		goto bad;
	}

	if (wdcprobe(chp, NULL) == 0)
		goto bad;

	wd->sc_chp = chp;
	if (whdattach(wd) != 0)
		goto bad;

	WDC_PCCS_DEBUG("wdc_pccs: IDE connected\n")

 	lp = wd->sc_dk.dk_label;
	da->da_cyl = lp->d_ncylinders;
	da->da_head = lp->d_ntracks;
	da->da_sec = lp->d_nsectors;
	return wd;

bad:	
	pccs_close(sid, ci);
	return NULL;
}

int
wdc_pccs_read(arg, blkno, nblks, bufp)
	void *arg;
	daddr_t blkno;
	daddr_t nblks;
	u_int8_t *bufp;
{
	struct wd_softc *wd = arg;
	struct disklabel *lp = wd->sc_dk.dk_label;
	daddr_t i;
	int error = 0;

	for (i = 0; i < nblks && error == 0; i ++)
	{
		error = whdread(wd, blkno, bufp);
		bufp += lp->d_secsize;
		blkno ++;
	}

	return error;
}
#endif	/* BOOT_PCCS_WDC */
