/*	$NetBSD$	*/
#ifdef	BOOT_WDC
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * DMA and multi-sector PIO handling are derived from code contributed by
 * Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles M. Hannum.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <sys/disk.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <machine/cpu.h>
#include "busio_user.h"
#include "wdcsubr.h"

/* local port defs (relocated) */
#define	wd_data		(0)
#define wd_error	(1)
#define	wd_precomp	(1)
#define	wd_features	(1)
#define	wd_seccnt	(2)
#define	wd_sector	(3)
#define	wd_cyl_lo	(4)
#define	wd_cyl_hi	(5)
#define	wd_sdh		(6)
#define	wd_command	(7)
#define	wd_status	(7)

#define	wd_base		(8)
#define	wd_altsts	(0)
#define	wd_ctlr		(0)
#define	wd_digin	(1)
#define	wd_mctlr	(1)

#define	wd_cridx	(2)
#define	wd_crdata	(3)
#define	wd_edata	(4)

/* bank select macro */
#define	RECOVERYTIME		(hz / 2) /* time to recover from an error */
#define	WDCDELAY		100
#define	WDCNDELAY		100000	 /* 10 sec */

#define	wait_for_drq(d)		wdcwait(d, WDCS_DRDY | WDCS_DSC | WDCS_DRQ)
#define	wait_for_ready(d)	wdcwait(d, WDCS_DRDY | WDCS_DSC)
#define	wait_for_unbusy(d)	wdcwait(d, 0)

/* common access methods */
#define	wdc_select_bank(chp, bank) 				\
{								\
	if ((chp)->ch_bank != 0)				\
		((*((chp)->ch_hwfuncs.banksel)) ((chp), (bank)));\
}
#define	wdc_cr0_read(chp, offs) 				\
	((*((chp)->ch_hwfuncs.cr0_read)) ((chp), (offs)))
#define	wdc_cr0_write(chp, offs, val) 				\
	((*((chp)->ch_hwfuncs.cr0_write)) ((chp), (offs), (val)))
#define	wdc_cr1_read(chp, offs) 				\
	((*((chp)->ch_hwfuncs.cr1_read)) ((chp), (offs)))
#define	wdc_cr1_write(chp, offs, val) 				\
	((*((chp)->ch_hwfuncs.cr1_write)) ((chp), (offs), (val)))
#define	wdc_data_read(chp, bufp, cnt, bus)			\
	((*((chp)->ch_hwfuncs.data_read)) ((chp), (bufp), (cnt), (bus)))
#define	wdc_data_write(chp, bufp, cnt, bus)			\
	((*((chp)->ch_hwfuncs.data_write)) ((chp), (bufp), (cnt), (bus)))

static int _wdcprobesubr __P((struct channel_softc *, int, int));
static int wdcreset __P((struct channel_softc *));
static void wdc_initialize_chip __P((struct channel_softc *));
static int wdccommand __P((struct channel_softc *, int, int, int, int, int, int));
static int wdccommandshort __P((struct channel_softc *, int, int));
static int wdcwait __P((struct channel_softc *, int));

static __inline u_int8_t wd_precv __P((struct wd_softc *));
static int wd_load_params __P((struct wd_softc *));
static int wdcontrol __P((struct wd_softc *));
static int wdc_inquire_idedrv __P((struct channel_softc *, int));
static void wdc_init_channel_softc __P((struct channel_softc *, struct wdc_hw *));

#define	WDCSUBR_DEBUG(mesg)	{printf(mesg);}
/*********************************************************
 * probe attach
 *********************************************************/
struct wdc_hw wdc_generic_hardware = {
	"ATbus",
	WDC_DTYPE_GENERIC,

	WDC_ACCESS_PLAIN,
	2,

	{ NULL,
	  NULL,
	  NULL,
	  wdc_cr0_read_plain,
	  wdc_cr0_write_plain,
	  wdc_cr1_read_plain,
	  wdc_cr1_write_plain,
	  wdc_data_read_generic,
	  wdc_data_write_generic
	},

	/* WDC_CAPABILITY_DATA16 | WDC_CAPABILITY_MODE */ 0,
	3,
	0
};

static void
wdc_init_channel_softc(chp, hwp)
	struct channel_softc *chp;
	struct wdc_hw *hwp;
{

	/*
	 * Setup hardware access methods
	 */
	if (hwp == NULL)
		hwp = &wdc_generic_hardware;

	chp->ch_hwfuncs = hwp->hw_funcs;
	chp->ch_hw = hwp;
	chp->ch_delay = WDCDELAY;

	/*
	 * Each bus should setup io and mem space handles.
	 */
}

static int
_wdcprobesubr(chp, drive, flags)
	struct channel_softc *chp;
	int drive, flags;
{
	int rv = 0;

	wdc_select_bank(chp, drive);
	if (wdc_cr0_read(chp, wd_status) == (u_int8_t) -1)
		goto out;

	if (wdcreset(chp) != 0)
		goto out;

	wdc_select_bank(chp, drive);
	wdc_cr0_write(chp, wd_sdh, WDSD_IBM | (WDC_DRIVE_NO(chp, drive) << 4));
	if (wait_for_unbusy(chp) < 0)
		goto out;

	if (wdc_inquire_idedrv(chp, drive) == ENXIO)
		goto out;
	rv = 1;

out:
	wdc_select_bank(chp, WDC_BANK_UNKNOWN);
	return rv;
}

int
wdcprobe(chp, hw)
	struct channel_softc *chp;
	void *hw;
{
	int rv;

	wdc_init_channel_softc(chp, hw);
	chp->data16iot = chp->cmd_iot;
	chp->data16ioh = chp->cmd_ioh;

	wdc_initialize_chip(chp);
	rv = _wdcprobesubr(chp, chp->channel, 0);
	return rv;
}

/*********************************************************
 * ide drive
 *********************************************************/
static int
wdc_inquire_idedrv(chp, unit)
	struct channel_softc *chp;
	int unit;
{
	u_int8_t regv;
	int error;

	/* select bank */
	wdc_select_bank(chp, unit);
	wdc_cr0_write(chp, wd_sdh, WDSD_IBM | (WDC_DRIVE_NO(chp, unit) << 4));

	regv = wdc_cr0_read(chp, wd_status);
	if ((regv & (WDCS_BSY | WDCS_DRQ | WDCS_CORR)) != 0)
		return ENXIO;

	wdc_cr0_write(chp, wd_command, WDCC_RECAL);
	delay(1000);		/* 1ms wait */

	regv = wdc_cr0_read(chp, wd_status);
	if ((regv & (WDCS_BSY | WDCS_DRDY | WDCS_DSC | WDCS_ERR)) == 0)
		return ENXIO;

	error = wait_for_unbusy(chp);
	if (error < 0)
		return ENXIO;
	if (error > 0)
	{
		delay(1000);	/* 1ms wait */
		return EIO;
	}
	error = wait_for_ready(chp);
	if (error < 0)
		return ENXIO;
	return error;
}	

int
wdc_load_params(chp, ap)
	struct channel_softc *chp;
	struct ataparams *ap;
{
	u_int8_t tbp[DEV_BSIZE + 2];

	memset(tbp, 0, DEV_BSIZE);
	if (wdccommandshort(chp, chp->channel, WDCC_IDENTIFY) != 0)
		return EIO;

	if (wait_for_drq(chp) != 0)
		return EIO;

	wdc_data_read(chp, tbp, DEV_BSIZE, 0);
	wait_for_ready(chp);
	bcopy(tbp, ap, sizeof(*ap));
	return 0;
}

/*********************************************************
 * chip command out
 *********************************************************/
static int
wdccommand(chp, command, cylin, head, sector, count, features)
	struct channel_softc *chp;
	int command;
	int cylin, head, sector, count, features;
{

	/* Select drive, head, and addressing mode. */
	wdc_select_bank(chp, chp->channel);
	wdc_cr0_write(chp, wd_sdh, WDSD_IBM | 
		      (WDC_DRIVE_NO(chp, chp->channel) << 4) | head);

	/* Wait for it to become ready to accept a command. */
	if (wdcwait(chp, WDCS_DRDY) < 0)
		return -1;

	/* Load parameters. */
	wdc_cr0_write(chp, wd_features, features);
	wdc_cr0_write(chp, wd_cyl_lo, cylin);
	wdc_cr0_write(chp, wd_cyl_hi, cylin >> 8);
	wdc_cr0_write(chp, wd_sector, sector);
	wdc_cr0_write(chp, wd_seccnt, count);

	/* Send command. */
	wdc_cr0_write(chp, wd_command, command);

	return 0;
}

static int
wdccommandshort(chp, chan, command)
	struct channel_softc *chp;
	int chan;
	int command;
{

	/* Select drive. */
	wdc_select_bank(chp, chan);
	wdc_cr0_write(chp, wd_sdh, WDSD_IBM | (WDC_DRIVE_NO(chp, chan) << 4));
	if ((command & WDCC_FORCE) == 0 && wdcwait(chp, WDCS_DRDY) < 0)
		return -1;

	wdc_cr0_write(chp, wd_command, command & WDCC_CMDMASK);
	return 0;
}

/*********************************************************
 * chip reset
 *********************************************************/
static int
wdcreset(chp)
	struct channel_softc *chp;
{

	wdc_cr1_write(chp, wd_ctlr, WDCTL_IDS);
	delay(1000);

	(void) wdc_cr0_read(chp, wd_error);

	wdc_cr1_write(chp, wd_ctlr, WDCTL_4BIT);
	if (wait_for_unbusy(chp) < 0)
		return 1;
	return 0;
}

/*********************************************************
 * chip phase wait
 *********************************************************/
static int
wdcwait(chp, mask)
	struct channel_softc *chp;
	int mask;
{
	int timeout = 0;
	u_int8_t status;

	for (;;)
	{
		chp->ch_status = status = wdc_cr0_read(chp, wd_status);
		if ((status & WDCS_BSY) == 0 && (status & mask) == mask)
			break;
		if (++ timeout > WDCNDELAY)
			return -1;
		delay(chp->ch_delay);
	}

	if (status & WDCS_ERR)
	{
		chp->ch_error = wdc_cr0_read(chp, wd_error);
		return WDCS_ERR;
	}

	return 0;
}

/*********************************************************
 * IDE bank control
 *********************************************************/
static void
wdc_initialize_chip(chp)
	struct channel_softc *chp;
{
	struct wdc_hw *hw = chp->ch_hw;

	if ((hw->hw_access & WDC_ACCESS_HASBANK) != 0 &&
	    chp->ch_hwfuncs.banksel != NULL)
		chp->ch_bank = 1;

	if (chp->ch_hwfuncs.open != NULL)
		(*chp->ch_hwfuncs.open) (chp);
}

#if	0
static void wdc_release_chip __P((struct channel_softc *));

static void
wdc_release_chip(chp)
	struct channel_softc *chp;
{

	chp->ch_bank = 0;
	if (chp->ch_hwfuncs.close != NULL)
		(*chp->ch_hwfuncs.close) (chp);
}
#endif

/*********************************************************
 * data access methods
 *********************************************************/
u_int8_t
wdc_cr0_read_plain(chp, offs)
	struct channel_softc *chp;
	bus_addr_t offs;
{

	return bus_space_read_1(chp->cmd_iot, chp->cmd_ioh, offs);
}


u_int8_t
wdc_cr0_read_indexed(chp, offs)
	struct channel_softc *chp;
	bus_addr_t offs;
{

	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, wd_cridx, offs);
	return bus_space_read_1(chp->cmd_iot, chp->cmd_ioh, wd_crdata);
}
					
void
wdc_cr0_write_plain(chp, offs, data)
	struct channel_softc *chp;
	bus_addr_t offs;
	u_int8_t data;
{

	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, offs, data);
}

void
wdc_cr0_write_indexed(chp, offs, data)
	struct channel_softc *chp;
	bus_addr_t offs;
	u_int8_t data;
{

	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, wd_cridx, (u_int8_t) offs);
	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, wd_crdata, data);
}

u_int8_t
wdc_cr1_read_plain(chp, offs)
	struct channel_softc *chp;
	bus_addr_t offs;
{

	return bus_space_read_1(chp->ctl_iot, chp->ctl_ioh, offs);
}


u_int8_t
wdc_cr1_read_indexed(chp, offs)
	struct channel_softc *chp;
	bus_addr_t offs;
{

	offs += wd_base;
	bus_space_write_1(chp->ctl_iot, chp->cmd_ioh, wd_cridx, offs);
	return bus_space_read_1(chp->ctl_iot, chp->cmd_ioh, wd_crdata);
}
					
void
wdc_cr1_write_plain(chp, offs, data)
	struct channel_softc *chp;
	bus_addr_t offs;
	u_int8_t data;
{

	bus_space_write_1(chp->ctl_iot, chp->ctl_ioh, offs, data);
}

void
wdc_cr1_write_indexed(chp, offs, data)
	struct channel_softc *chp;
	bus_addr_t offs;
	u_int8_t data;
{

	offs += wd_base;
	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, wd_cridx, offs);
	bus_space_write_1(chp->cmd_iot, chp->cmd_ioh, wd_crdata, data);
}

/*********************************************************
 * data access methods (multi)
 *********************************************************/
/* generic 16 or 32 bits multi read */
void
wdc_data_read_generic(chp, bufp, cnt, flags)
	struct channel_softc *chp;
	u_int8_t *bufp;
	u_int cnt;
	int flags;
{

	bus_space_read_multi_2(chp->data16iot, chp->data16ioh,
		wd_data, (u_int16_t *) bufp, cnt >> 1);
}

void
wdc_data_write_generic(chp, bufp, cnt, flags)
	struct channel_softc *chp;
	u_int8_t *bufp;
	u_int cnt;
	int flags;
{

	bus_space_write_multi_2(chp->data16iot, chp->data16ioh,
		wd_data, (u_int16_t *) bufp, cnt >> 1);
}

/*******************************************************
 * WD control
 *******************************************************/
#ifdef	USER_SPACE_DEBUG
static void wd_print_params __P((struct wd_softc *));
#endif /* USER_SPACE_DEBUG */

int
whdattach(wd)
	struct wd_softc *wd;
{

	wd->sc_dk.dk_label = &wd->sc_label;
	if (wd_load_params(wd) != 0)
		return EINVAL;
	if (wdcontrol(wd) != 0)
		return EINVAL;
#ifdef	USER_SPACE_DEBUG
	wd_print_params(wd);
#endif /* USER_SPACE_DEBUG */
	return 0;
}

#ifdef	USER_SPACE_DEBUG
static void
wd_print_params(wd)
	struct wd_softc *wd;
{
	int i, blank;
	char buf[41], c, *p, *q;

	for (blank = 0, p = wd->sc_params.atap_model, q = buf, i = 0;
	     i < sizeof(wd->sc_params.atap_model); i++)
	{
		c = *p++;
		if (c == '\0')
			break;
		if (c != ' ') {
			if (blank) {
				*q++ = ' ';
				blank = 0;
			}
			*q++ = c;
		} else
			blank = 1;
	}
	*q++ = '\0';

	printf("%s: %dMB, %d cyl, %d head, %d sec, %d bytes/sec <%s>\n",
		"wd",
		wd->sc_params.atap_cylinders *
		(wd->sc_params.atap_heads * wd->sc_params.atap_sectors) /
		(1048576 / DEV_BSIZE),
		wd->sc_params.atap_cylinders,
		wd->sc_params.atap_heads, wd->sc_params.atap_sectors,
	    	DEV_BSIZE, buf);
}
#endif	/* USER_SPACE_DEBUG */

/*
 * Read/write routine for a buffer.  Validates the arguments and schedules the
 * transfer.  Does not wait for the transfer to complete.
 */
static __inline u_int8_t
wd_precv(wd)
	struct wd_softc *wd;
{
	struct disklabel *lp = wd->sc_dk.dk_label;

	return ((lp->d_type == DTYPE_ST506) ? (lp->d_precompcyl / 4) : 0);
}

static int
wd_load_params(wd)
	struct wd_softc *wd;
{
	struct channel_softc *chp = wd->sc_chp;
	struct disklabel *lp = wd->sc_dk.dk_label;
	int error;

	error = wdc_load_params(chp, &wd->sc_params);
	if (error != 0)
	{
		/*
		 * We `know' there's a drive here; just assume it's old.
		 * This geometry is only used to read the MBR and print a
		 * (false) attach message.
		 */
		strncpy(wd->sc_dk.dk_label->d_typename, "ST506",
		    sizeof wd->sc_dk.dk_label->d_typename);
		wd->sc_dk.dk_label->d_type = DTYPE_ST506;

		strncpy(wd->sc_params.atap_model, "unknown",
			sizeof(wd->sc_params.atap_model));
		wd->sc_params.atap_config = ATA_CFG_FIXED;
		wd->sc_params.atap_cylinders = 1024;
		wd->sc_params.atap_heads = 8;
		wd->sc_params.atap_sectors = 17;
		wd->sc_params.atap_multi = 0;
		wd->sc_params.atap_capabilities1 = 0;
	}
	else
	{
		/* For over 8.4GB HDD */
		if (((wd->sc_params.atap_capabilities1 & WDC_CAP_LBA) != 0) &&
		    (wd->sc_params.atap_cylinders >= 16383))
		{
			u_int32_t totalsecs, secspercyl;

		    	totalsecs = wd->sc_params.atap_capacity[1];
			totalsecs <<= 16;
		    	totalsecs += wd->sc_params.atap_capacity[0];
			secspercyl = wd->sc_params.atap_heads *
				     wd->sc_params.atap_sectors;
			if (secspercyl == 0)
				secspercyl = 17 * 8;	/* XXX */
			wd->sc_params.atap_cylinders = totalsecs / secspercyl;
		}
		else
		{
			u_int32_t totalsecs, secspercyl;

			secspercyl = wd->sc_params.atap_heads;
			secspercyl *= wd->sc_params.atap_sectors;
			totalsecs = wd->sc_params.atap_cylinders;
			totalsecs *= secspercyl;
			wd->sc_params.atap_heads = 8;
			wd->sc_params.atap_sectors = 17;
			wd->sc_params.atap_cylinders = totalsecs / (8 * 17);
		}
	
		strncpy(wd->sc_dk.dk_label->d_typename, "ESDI/IDE",
		    sizeof(wd->sc_dk.dk_label->d_typename));
		wd->sc_dk.dk_label->d_type = DTYPE_ESDI;
	}

	memset(lp, 0, sizeof(struct disklabel));
	lp->d_secsize = DEV_BSIZE;
	lp->d_ntracks = wd->sc_params.atap_heads;
	lp->d_nsectors = wd->sc_params.atap_sectors;
	lp->d_ncylinders = wd->sc_params.atap_cylinders;
	lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;
	lp->d_type = DTYPE_ST506;
	lp->d_subtype = chp->ch_hw->d_subtype;
	strncpy(lp->d_typename, "ST506 disk", 16);
	strncpy(lp->d_packname, wd->sc_params.atap_model, 16);
	lp->d_secperunit = lp->d_secpercyl * lp->d_ncylinders;
	lp->d_rpm = 3600;
	lp->d_interleave = 1;
	lp->d_flags = 0;

	lp->d_partitions[RAW_PART].p_offset = 0;
	lp->d_partitions[RAW_PART].p_size = lp->d_secperunit;
	lp->d_partitions[RAW_PART].p_fstype = FS_UNUSED;
	lp->d_npartitions = RAW_PART + 1;
	return 0;
}

int
whdread(wd, blkno, bufp)
	struct wd_softc *wd;
	daddr_t blkno;
	u_int8_t *bufp;
{
	struct channel_softc *chp = wd->sc_chp;
	struct disklabel *lp = wd->sc_dk.dk_label;
	int nbytes, nblks, sector, cylin, head;

	nblks = 1;
	nbytes = lp->d_secsize * nblks;
	if ((wd->sc_params.atap_capabilities1 & WDC_CAP_LBA) != 0)
	{
		sector = (blkno >> 0) & 0xff;
		cylin = (blkno >> 8) & 0xffff;
		head = (blkno >> 24) & 0xf;
		head |= WDSD_LBA;
	}
	else
	{
		sector = blkno % lp->d_nsectors;
		sector++;	/* Sectors begin with 1, not 0. */
		blkno /= lp->d_nsectors;
		head = blkno % lp->d_ntracks;
		blkno /= lp->d_ntracks;
		cylin = blkno;
		head |= WDSD_CHS;
	}

	if (wdccommand(chp, WDCC_READ, cylin, head, sector, nblks,
		       wd_precv(wd)) != 0)
		return EIO;

	if (wait_for_unbusy(chp) < 0)
		return EIO;

	if (wait_for_drq(chp) != 0)
		return EIO;

	if ((chp->ch_status & (WDCS_DRDY | WDCS_DSC | WDCS_DRQ))
		    != (WDCS_DRDY | WDCS_DSC | WDCS_DRQ))
		return EIO;

	wdc_data_read(chp, (u_int8_t *) bufp, nbytes, 0);
	return 0;
}

/*******************************************************
 * drive control
 *******************************************************/
static int
wdcontrol(wd)
	struct wd_softc *wd;
{
	struct channel_softc *chp = wd->sc_chp;

	wdccommandshort(chp, chp->channel, WDCC_RECAL | WDCC_FORCE);
	wait_for_unbusy(chp);

	if (wdccommandshort(chp, chp->channel, WDCC_RECAL) != 0)
		return EIO;
	if (wait_for_unbusy(chp) < 0)
		return EIO;
	if (chp->ch_status & (WDCS_ERR | WDCS_DWF))
		return EIO;

	if (wdccommand(chp, WDCC_IDP, wd->sc_dk.dk_label->d_ncylinders,
	    wd->sc_dk.dk_label->d_ntracks - 1, 0,
	    wd->sc_dk.dk_label->d_nsectors, wd_precv(wd)) != 0)
		return EIO;
	if (wait_for_unbusy(chp) < 0)
		return EIO;
	if (chp->ch_status & (WDCS_ERR | WDCS_DWF))
		return EIO;
	return 0;
}
#endif	/* BOOT_WDC */
