/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 2000
 *	NetBSD/pc98 porting staff. All rights reserved.
 *  Copyright (c) 2000
 *	Naofumi HONDA. All rights reserved.
 */
/*
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * DMA and multi-sector PIO handling are derived from code contributed by
 * Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles M. Hannum.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <i386/Cbus/dev/atapi/wdcreg.h>
#include <dev/ata/atareg.h>

#define	ATABUS_DRIVE_PER_BANK 2
#define	ATABUS_MAXDRIVE	4

#define	WDC_DRIVE_NO(chp, chan)	((chan) % ATABUS_DRIVE_PER_BANK)
#define	WDC_BANK_NO(chp, chan)	((chan) / ATABUS_DRIVE_PER_BANK)
#define	WDC_BANK_TO_CHAN(bank, drive) ((bank) * ATABUS_DRIVE_PER_BANK + (drive))

#define	WDC_SECOND_BANK		ATABUS_DRIVE_PER_BANK
#define	WDC_BANK_UNKNOWN	(-1)

struct channel_softc;
struct wd_softc;
struct wdc_hwfuncs {
	void (*open) __P((struct channel_softc *));
	void (*close) __P((struct channel_softc *));

	void (*banksel) __P((struct channel_softc *, u_int));

	u_int8_t (*cr0_read) __P((struct channel_softc *, bus_addr_t));
	void (*cr0_write) __P((struct channel_softc *, bus_addr_t, u_int8_t));

	u_int8_t (*cr1_read) __P((struct channel_softc *, bus_addr_t));
	void (*cr1_write) __P((struct channel_softc *, bus_addr_t, u_int8_t));

	void (*data_read) __P((struct channel_softc *, u_int8_t *, u_int, int));
	void (*data_write) __P((struct channel_softc *, u_int8_t *, u_int, int));
};

struct wdc_hw {
	u_char *hw_name;			/* symbolic name */
#define	WDC_DTYPE_GENERIC	0
#define	WDC_DTYPE_IDE98		1		/* XXX */
	u_int d_subtype;			/* geometric subtype */

#define	WDC_ACCESS_PLAIN	0x0100
#define	WDC_ACCESS_INDEXED	0x0200
#define	WDC_ACCESS_MEM		0x0400
#define	WDC_ACCESS_HASBANK	0x0800
#define	WDC_ACCESS_USER		0x00ff
#define	WDC_ACCESS_BRESET	0x0001
	int hw_access;				/* access type */
	int hw_maxunit;				/* max unit */

	struct wdc_hwfuncs hw_funcs;		/* access methods */

	int hw_capability;			/* capability */
	int hw_pio;
	int hw_dma;
}; 

struct channel_softc {
	int channel;

	bus_space_tag_t cmd_iot;
	bus_space_tag_t ctl_iot;
	bus_space_tag_t data16iot;
	bus_space_tag_t data32iot;

	bus_space_handle_t cmd_ioh;
	bus_space_handle_t ctl_ioh;
	bus_space_handle_t data16ioh;
	bus_space_handle_t data32ioh;

	bus_space_tag_t ch_bkbst;
	bus_space_handle_t ch_bkbsh;

	bus_space_tag_t ch_memt;
	bus_space_handle_t ch_memh;
	vsize_t ch_vsize;

	int ch_delay;
	int ch_bank;			/* has bank switch register */
	int ch_errors;			/* errors during current transfer */
	u_int8_t ch_status;		/* copy of status register */
	u_int8_t ch_error;		/* copy of error register */

#define	MODE_IDE	0
#define	MODE_ATAPI	1
#define	MODE_WATAPI	2
	u_int ch_interface;		/* current interface mode */
	struct wdc_hw *ch_hw;		/* port select */
	struct wdc_hwfuncs ch_hwfuncs;	/* hw interface functions */
};

struct wd_softc {
	struct disk sc_dk;
	struct channel_softc *sc_chp;

	struct disklabel sc_label;
	struct ataparams sc_params;
};

int wdcprobe __P((struct channel_softc *, void *));
int whdattach __P((struct wd_softc *));
int whdread __P((struct wd_softc *, daddr_t, u_int8_t *));

/* methods */
u_int8_t wdc_cr0_read_plain __P((struct channel_softc *, bus_addr_t));
void wdc_cr0_write_plain __P((struct channel_softc *, bus_addr_t, u_int8_t));
u_int8_t wdc_cr0_read_indexed __P((struct channel_softc *, bus_addr_t));
void wdc_cr0_write_indexed __P((struct channel_softc *, bus_addr_t, u_int8_t));
u_int8_t wdc_cr1_read_plain __P((struct channel_softc *, bus_addr_t));
void wdc_cr1_write_plain __P((struct channel_softc *, bus_addr_t, u_int8_t));
u_int8_t wdc_cr1_read_indexed __P((struct channel_softc *, bus_addr_t));
void wdc_cr1_write_indexed __P((struct channel_softc *, bus_addr_t, u_int8_t));
void wdc_data_write_generic __P((struct channel_softc *, u_int8_t *, u_int, int));
void wdc_data_read_generic __P((struct channel_softc *, u_int8_t *, u_int, int));


/* XXX */
int	 strncmp __P((const char *, const char *, size_t));
char	*strncpy __P((char *, const char *, size_t));
