/*	$NetBSD$	*/
/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1999 NetBSD/pc98 porting staff.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <sys/buf.h>
#include <nlist.h>
#include <kvm.h>

#define	_KERNEL
#include <machine/bootinfo.h>
#include <machine/bootflags.h>

#include "keycap.h"
static int parse_conf __P((struct bootflags_device *));
static u_char * getstring __P((u_char *, u_char *));
static int make_db __P((u_char *, u_char *));
static int save_db __P((u_char *, u_char *));
static int __dump_db __P((FILE *, struct bootflags_info_header *));
static int dump_db __P((u_char *, u_char *));
static void __dump_val __P((FILE *, u_char *, u_long));

u_char *myname;

/*******************************************************
 * MACRO
 *******************************************************/
#define	MAX_CONFFILE_SIZE 0x10000
#define CAPBUFSZ 1024
#define	DBBUFSZ 1024

static char dbbuf[DBBUFSZ];

/*******************************************************
 * MISC
 *******************************************************/
#define	FLAGSET(name, ref, val)				\
	{						\
		if (kgetflag((name))) 			\
			(ref) |= (val);			\
	 }

#define	VALSET(name, ref) 				\
	{ 						\
		register int tmp = kgetnum((name)); 	\
							\
		if (tmp != -1) 				\
			(ref) = tmp;			\
	}

static u_char *
getstring(name, ref)
	u_char *name, *ref;
{
	char *pos, **str;

	*ref = 0;
	pos = ref;
	str = &pos;
	return kgetstr(name, str);
}

static int
parse_conf(bdp)
	struct bootflags_device *bdp;
{
	u_char tbuf[CAPBUFSZ];
	u_char buf[CAPBUFSZ];
	int i, rv;

	bzero(buf, sizeof(buf));
	rv = kgetentnext(buf);
	if (rv <= 0)
		return ENOENT;

	ksetup(buf);
	buf[CAPBUFSZ - 1] = 0;

	bzero(bdp->bd_name, sizeof(bdp->bd_name));
	if (getstring("name", tbuf) == NULL)
	{
		if (getstring("dv", tbuf) == NULL)	/* old compat */
			return EINVAL;
	}
	strncpy(bdp->bd_name, tbuf, sizeof(bdp->bd_name) - 1);

	for (i = 0; i < 8; i ++)
		bdp->bd_loc[i] = -1;

	bdp->bd_type = 0;
	if (getstring("bus", tbuf) != NULL)
	{
		if (strstr(tbuf, "boot") != NULL)
		{
			bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT;
		}

		if (strstr(tbuf, "isa") != NULL)
		{
			bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_BUS_ISA;
		}

		if (strstr(tbuf, "disk") != NULL)
		{
			bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_BUS_DISK;
		}

		if (strstr(tbuf, "pci") != NULL)
		{
			bdp->bd_type |= BOOTFLAGS_DEVICE_TYPE_BUS_PCI;
		}
	}

	if ((bdp->bd_type & BOOTFLAGS_DEVICE_TYPE_TARGET_MASK) ==
	     BOOTFLAGS_DEVICE_TYPE_TARGET_KERNEL)
	{
		switch (bdp->bd_type & BOOTFLAGS_DEVICE_TYPE_BUS_MASK)
		{
		case BOOTFLAGS_DEVICE_TYPE_BUS_ISA:
			VALSET("sel", bdp->bd_loc[0]);
			VALSET("io", bdp->bd_loc[2]);
			VALSET("iomem", bdp->bd_loc[4]);
			VALSET("iosiz", bdp->bd_loc[5]);
			VALSET("irq", bdp->bd_loc[6]);
			VALSET("drq", bdp->bd_loc[7]);
			bdp->bd_flags = -1;
			VALSET("flags", bdp->bd_flags);
			return 0;

		case BOOTFLAGS_DEVICE_TYPE_BUS_PCI:
			VALSET("sel", bdp->bd_loc[0]);
			bdp->bd_flags = -1;
			VALSET("flags", bdp->bd_flags);
			return 0;

		case BOOTFLAGS_DEVICE_TYPE_BUS_DISK:
			VALSET("sel", bdp->bd_loc[0]);
			VALSET("ntracks", bdp->bd_loc[1]);
			VALSET("nsectors", bdp->bd_loc[2]);
			return 0;
		}
	}

	if ((bdp->bd_type & BOOTFLAGS_DEVICE_TYPE_TARGET_MASK) == 
	     BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT)
	{
		if (strstr(bdp->bd_name, "eth.addr") != NULL)
		{
			VALSET("addr0", bdp->bd_loc[0]);
			VALSET("addr1", bdp->bd_loc[1]);
			VALSET("addr2", bdp->bd_loc[2]);
			VALSET("addr3", bdp->bd_loc[3]);
			VALSET("addr4", bdp->bd_loc[4]);
			VALSET("addr5", bdp->bd_loc[5]);
			return 0;
		}

		if ((bdp->bd_type & BOOTFLAGS_DEVICE_TYPE_BUS_MASK) ==
		     BOOTFLAGS_DEVICE_TYPE_BUS_ISA)
		{
			VALSET("sel", bdp->bd_loc[0]);
			VALSET("io", bdp->bd_loc[2]);
			VALSET("iomem", bdp->bd_loc[4]);
			VALSET("iosiz", bdp->bd_loc[5]);
			bdp->bd_flags = -1;
			VALSET("flags", bdp->bd_flags);
			return 0;
		}
	}
	return 0;
}

/*******************************************************
 * Dump data
 *******************************************************/
static void
__dump_val(fp, s, val)
	FILE *fp;
	u_char *s;
	u_long val;
{

	if (val != (u_long) -1)
		fprintf(fp, "\t:%s#0x%x:\\\n", s, val);
}

static int
dump_db(data_file, db_file)
	u_char *data_file;
	u_char *db_file;
{
	struct bootflags_info_header *bip;
	FILE *fp;
	int rv;

	bip = (void *) dbbuf;
	fp = fopen(db_file, "r");
	if (fp == NULL)
		return EIO;
	if (fread(dbbuf, 1, DBBUFSZ, fp) < sizeof(*bip))
		return EIO;
	fclose(fp);

	fp = fopen(data_file, "w");
	if (fp == NULL)
		return EIO;
	rv = __dump_db(fp, bip);
	fclose(fp);
	return rv;
}

static int
__dump_db(fp, bip)
	FILE *fp;
	struct bootflags_info_header *bip;
{
	struct bootflags_device *bdp;
	int i, j, nent;
	u_char *s;

	bdp = (void *) ((u_int8_t *) bip + sizeof(*bip));
	if (bip->bi_signature != BOOTFLAGS_SIGNATURE)
		return EINVAL;
	nent = bip->bi_nentry;

	fprintf(fp, "# NetBSD/pc98\n");
	fprintf(fp, "# bootflags.conf is automatically generated\n#\n");
	for (i = 0; i < nent; i ++, bdp ++)
	{
		fprintf(fp, "bootflags entry %d|kernel settings:\\\n", i);
		switch (bdp->bd_type)
		{
			case BOOTFLAGS_DEVICE_TYPE_TARGET_KERNEL |
			     BOOTFLAGS_DEVICE_TYPE_BUS_PCI:
				s = "kernel.pci";
				break;
			case BOOTFLAGS_DEVICE_TYPE_TARGET_KERNEL |
			     BOOTFLAGS_DEVICE_TYPE_BUS_DISK:
				s = "kernel.disk";
				break;
			default:	
			case BOOTFLAGS_DEVICE_TYPE_TARGET_KERNEL |
			     BOOTFLAGS_DEVICE_TYPE_BUS_ISA:
				s = "kernel.isa";
				break;
			case BOOTFLAGS_DEVICE_TYPE_TARGET_BOOT |
			     BOOTFLAGS_DEVICE_TYPE_BUS_ISA:
				s = "boot.isa";
				break;
		}

		fprintf(fp, "\t:bus=%s:\\\n", s);
		fprintf(fp, "\t:name=%s:\\\n", bdp->bd_name);
		__dump_val(fp, "flags", bdp->bd_flags);

		if (strstr(bdp->bd_name, ".eth.addr") != NULL)
		{
			__dump_val(fp, "addr0", bdp->bd_loc[0]);
			__dump_val(fp, "addr1", bdp->bd_loc[1]);
			__dump_val(fp, "addr2", bdp->bd_loc[2]);
			__dump_val(fp, "addr3", bdp->bd_loc[3]);
			__dump_val(fp, "addr4", bdp->bd_loc[4]);
			__dump_val(fp, "addr5", bdp->bd_loc[5]);
		}
		else 
		{
		switch (bdp->bd_type & BOOTFLAGS_DEVICE_TYPE_BUS_MASK)
		{
		case BOOTFLAGS_DEVICE_TYPE_BUS_ISA:
			__dump_val(fp, "sel", bdp->bd_loc[0]);
			__dump_val(fp, "io", bdp->bd_loc[2]);
			__dump_val(fp, "iomem", bdp->bd_loc[4]);
			__dump_val(fp, "iosiz", bdp->bd_loc[5]);
			__dump_val(fp, "irq", bdp->bd_loc[6]);
			__dump_val(fp, "drq", bdp->bd_loc[7]);
			break;
		case BOOTFLAGS_DEVICE_TYPE_BUS_PCI:
			__dump_val(fp, "sel", bdp->bd_loc[0]);
			break;
		case BOOTFLAGS_DEVICE_TYPE_BUS_DISK:
			__dump_val(fp, "sel", bdp->bd_loc[0]);
			__dump_val(fp, "ntracks", bdp->bd_loc[1]);
			__dump_val(fp, "nsectors", bdp->bd_loc[2]);
			break;
		}
		}
		fprintf(fp, "\n");
	}
	return 0;
}

/*******************************************************
 * Make bootflags.db from bootflags.conf
 *******************************************************/
static int
make_db(data_file, db_file)
	u_char *data_file, *db_file;
{
	int n, fd, ofd;
	struct stat file_stat;
	u_char *buf;
	u_int filesize, bufsize;
	u_int8_t *bp;
	struct bootflags_info_header *bip;
	struct bootflags_device *bdp;

	fd = open(data_file, O_RDONLY, 0);
	if (fd < 0)
	{
		perror(data_file);
		goto bad;
	}

	if (fstat(fd, &file_stat) != 0)
		goto bad;

	filesize = (u_int) file_stat.st_size;
	if (filesize > MAX_CONFFILE_SIZE)
		goto bad;

	bufsize = (filesize / 1024 + 2) * 1024;
	buf = malloc(bufsize);
	if (buf == NULL)
		goto bad;

	bzero(buf, bufsize);
	lseek(fd, (off_t) 0, 0);
	if (read(fd, buf, filesize) != filesize)
	{
		free(buf);
		goto bad;
	}

	ConfBuf = buf;
	ConfBufSize = filesize;
	close(fd);

	bp = malloc(sizeof(*bip) + sizeof(*bdp) * BOOTFLAGS_BTINFO_MAXENTRY);
	bip = (void *) bp;
	bdp = (void *) (bp + sizeof(*bip));
	bip->bi_signature = BOOTFLAGS_SIGNATURE;
	bip->bi_nentry = 0;

	kinit();
	for (n = 0; n < BOOTFLAGS_BTINFO_MAXENTRY; n ++, bdp ++)
	{
		if (parse_conf(bdp) == 0)
			bip->bi_nentry ++;
		else
			break;
	}

	fd = open(db_file, O_RDWR | O_CREAT | O_TRUNC, 0);
	if (fd < 0)
	{
		perror(db_file);
		goto bad;
	}
	write(fd, bp, sizeof(*bip) + sizeof(*bdp) * n);
	close(fd);
	return 0;

bad:
	return EIO;
}

/*******************************************************
 * Make bootflags.db from kernel data
 *******************************************************/
struct nlist namelist[] = {
	{ "_kernel_bootflags_table" },
	{ "" },
};

static int
save_db(data_file, db_file)
	u_char *data_file, *db_file;
{
	int fd, kfd, nent, size;
	u_char *buf;
	u_long toff;
	struct bootflags_btinfo_header *bbhp;
	struct bootflags_info_header *bip;
	struct bootflags_device *bdp;
	kvm_t *kd;

        kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, NULL);
	if (kd == 0) 
		return EIO;
	if (kvm_nlist(kd, namelist) != 0) 
		return EIO;
	if (kvm_read(kd, namelist[0].n_value, &toff, sizeof(toff)) != 
	    sizeof(toff))
		return EIO;
	kvm_close(kd);

	if (toff == 0)
		return ENOENT;

	kfd = open("/dev/kmem", O_RDONLY, 0);
	if (kfd < 0)
	{
		perror(myname);
		return EIO;
	}

	bbhp = (void *) dbbuf;
	bdp = (void *) ((u_int8_t *) dbbuf + sizeof(*bbhp));
	bip = &bbhp->bb_bi;

	lseek(kfd, (off_t) toff, 0);
	if (read(kfd, dbbuf, sizeof(*bbhp)) != sizeof(*bbhp))
		return EIO;
	
	if (bip->bi_signature != BOOTFLAGS_SIGNATURE)
		return EINVAL;

	nent = bip->bi_nentry;
	if (nent > BOOTFLAGS_BTINFO_MAXENTRY)
		return EINVAL;

	size = sizeof(*bbhp) + sizeof(*bdp) * nent;
	lseek(kfd, (off_t) toff, 0);
	if (read(kfd, dbbuf, size) != size)
		return EIO;
	close(kfd);

	fd = open(db_file, O_RDWR | O_CREAT | O_TRUNC, 0);
	if (fd < 0)
	{
		perror(db_file);
		return EIO;
	}
	write(fd, (u_int8_t *) bip, sizeof(*bip));
	write(fd, (u_int8_t *) bdp, sizeof(*bdp) * nent);
	close(fd);
	printf("%s: %s updated\n", myname, db_file);

	return 0;
}

/*******************************************************
 * Main
 *******************************************************/
int
main(argc, argv)
	int argc;
	char **argv;
{
	int ch, saveflags, dumpflags, rv;
        extern char *optarg;
        extern int optind;
        extern int optopt;
        extern int opterr;
	u_char *rootpath = "/etc";
	u_char data_file[1024];
	u_char db_file[1024];

	myname = argv[0];
	saveflags = dumpflags = 0;
	while ((ch = getopt(argc, argv, "f:sd")) != -1)
        {
                switch (ch)
                {
		case 'f':
			rootpath = optarg;
			break;
		case 's':
			saveflags = 1;
			break;
		case 'd':
			dumpflags = 1;
			break;
                default:
			continue;
                }
        }

	argc -= optind;
	argv += optind;

	sprintf(data_file, "%s/bootflags.conf", rootpath);
	sprintf(db_file, "%s/bootflags.db", rootpath);

	if (dumpflags != 0)
		rv = dump_db("/dev/stdout", db_file);
	else if (saveflags == 0)
		rv = make_db(data_file, db_file);
	else
		rv = save_db(data_file, db_file);

	return rv;
}
