/*	$NecBSD: diskinfo.c,v 1.6.10.2 2000/07/13 08:28:47 honda Exp $	*/
/*	$NetBSD$	*/
#define	DISKINFO_HAS_VPDUTIL

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1996, 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/disklabel.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/dkio.h>
#include <errno.h>
#include <machine/vscio.h>	/* XXX */
#include <machine/diskutil.h>

#define	BBSIZE	8192

struct diskinfo {
	u_char *name;
	u_int maxunit;
} diskinfo[] = { {"wd", 4}, {"sd", 7}, {NULL, 0} };

struct disklabel dl;

int check_rootdev __P((u_char *));
int check_target __P((u_char *, u_char *));
int main __P((int, u_char **));

int
main(argc, argv)
	int argc;
	u_char **argv;
{
	int unit;
	struct diskinfo *di;
	u_char devname[FILENAME_MAX + 1];
#ifdef	DISKINFO_HAS_VPDUTIL
	u_char cmdbuf[FILENAME_MAX + 1];
	u_char *vpd_device_name = NULL;
#endif	/* DISKINFO_HAS_VPDUTIL */

	printf("%6s %8s %8s %8s %8s %8s %10s\n",
		"disk", "bsize", "sectors", "heads", "cyls", "total", "size  ");
	di = &diskinfo[0];
	while (di->name != NULL)
	{
		for (unit = 0; unit < di->maxunit; unit++)
		{
			sprintf(devname, "%s%d", di->name, unit);
			check_target(devname, NULL);
			
		}
		di ++;
	}

#ifdef	DISKINFO_HAS_VPDUTIL
	vpd_device_name = NULL;
	if (argc == 2)
	{
		vpd_device_name = argv[1];
	}

	if (vpd_device_name != 0)
	{
		int slno;

		printf("\n");
		fflush(stdin);
		printf("use vpd device %s (y/n): ", vpd_device_name);
		if (getchar() != 'y')
			exit(0);

		fflush(stdin);
		strcpy(devname, "wd0");
		printf("target disk [%s]: ", devname);
		scanf("%s", devname);

		fflush(stdin);
		slno = 0;
		printf("target slice [%d]: ", slno);
		scanf("%d", &slno);

		sprintf(cmdbuf, "/sbin/vpdconfig %s %s %d\n",
			vpd_device_name, devname, slno);
		printf(cmdbuf);
		system(cmdbuf);
	}
#endif	/* DISKINFO_HAS_VPDUTIL */
	
	return 0;
}


int
check_target(dv, req)
	u_char *dv;
	u_char *req;
{
	struct disklabel *lp = &dl;
	int fd, ndospart, i,j;
	u_long scyl, ecyl;
	u_char devname[FILENAME_MAX + 1];
	u_char buf[BBSIZE];
	struct dos_partition *dp;
	float size;

	sprintf(devname, "/dev/r%sd", dv);
	if ((fd = open(devname, O_RDONLY, 0777)) < 0)
		return ENODEV;

	if (ioctl(fd, DIOCGDINFO, lp) < 0)
	{
bad:
		close(fd);
		return EIO;
	}

	scyl = disk_netbsd_start_sector(lp) / lp->d_secpercyl;
	ecyl = (disk_netbsd_start_sector(lp) + disk_netbsd_size_sector(lp)) /
		lp->d_secpercyl - 1;

	size = lp->d_secperunit / (1024 * 1024 / lp->d_secsize);
	printf("%6s %8d %8d %8d %8d %8d %8.1fMB\n",
		dv,
		lp->d_secsize,
		lp->d_nsectors,
		lp->d_ntracks,
		lp->d_ncylinders,
		lp->d_secperunit,
		(float) size);

	if (lseek(fd, (off_t) 0, 0) != (off_t) 0)
		goto bad;

	if (read(fd, buf, sizeof(buf)) != sizeof(buf))
		goto bad;

	dp = (struct dos_partition *) (buf + lp->d_secsize);
	ndospart = lp->d_secsize / 32;
	for (i = 0; i < ndospart; i ++, dp ++)
	{
		if (dos_partition_pc98_read_sysid(dp) == 0)
			continue;
		for (j = 0; j < 16; j ++)
			if (dp->dp_name[j] == ' ')
				dp->dp_name[j] = 0;
		printf("%6d%8x%8ld%8ld  %-16s\n",
			i, 
			(int) dos_partition_pc98_read_sysid(dp),
			disk_start_sector(lp, dp),
			disk_size_sector(lp, dp),
			dp->dp_name);
	}
	close(fd);
	return 0;
}
