/*	$NecBSD: vpdconfig.c,v 1.8.6.2.2.1 2001/04/15 23:18:54 honda Exp $	*/
/*	$NetBSD: vnconfig.c,v 1.8 1995/12/30 18:14:03 thorpej Exp $	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1997, 1998
 *	NetBSD/pc98 porting staff. All rights reserved.
 */
/*
 * Copyright (c) 1993 University of Utah.
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: vnconfig.c 1.1 93/12/15$
 *
 *	@(#)vnconfig.c	8.1 (Berkeley) 12/15/93
 */

#define	FSTYPENAMES
#define	DKTYPENAMES
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/disk.h>
#include <sys/disklabel.h>
#include <errno.h>
#include <machine/vpdio.h>

#define	BUFSIZE	1024

int main __P((int, char **));
static int decompose __P((u_char *));
static void usage __P((void));
static int show_info __P((char *, int));

int
main(argc, argv)
	int argc;
	char **argv;
{
	int verbose = 0, delete = 0, setup = 0;
	int ch, rv, fd;
	char *myprog;
	u_char buf[BUFSIZE];
	struct vpd_ioctl vpdio;
	int unit;

	myprog = argv[0];
	while ((ch = getopt(argc, argv, "vdh")) != -1)
	{
		switch (ch)
		{
		case 'v':
			verbose = 1;
			break;
		case 'd':
			delete = 1;
			break;

		case 'h':
		default:
		case '?':
			usage();
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (verbose != 0 || delete != 0)
	{
		if (argc != 1)
		{
			usage();
			exit(1);
		}
	}
	else 
	{
		switch (argc)
		{
		case	1:
			verbose = 1;
			break;

		case	3:
			setup = 1;
			break;
		
		default:
			usage();
			exit(1);
		}
	}

	strncpy(buf, argv[0], BUFSIZE);
	unit = decompose(buf);
	if (verbose == 0)
	{
		sprintf(buf, "/dev/rvpd%dc", unit);
	}
	else
	{
		sprintf(buf, "/dev/rvpd%dd", unit);
	}

	fd = open(buf, O_RDONLY, 0);
	if (fd < 0)
	{
		perror(buf);
		exit(1);
	}

	if (verbose != 0)
	{
		return show_info(argv[0], fd);
	}

	bzero(&vpdio, sizeof(vpdio));
	if (delete == 0)
	{
		strncpy(buf, argv[1], BUFSIZE);
		unit = decompose(buf);
		strncpy(vpdio.vpd_xname, buf, sizeof(vpdio.vpd_xname));
		vpdio.vpd_unit = unit;
		vpdio.vpd_flags = 0;
		vpdio.vpd_slno = atoi(argv[2]);
	}

	rv = ioctl(fd, VPDIOCSET, &vpdio);
	if (rv < 0)
	{
		perror(myprog);
		exit(1);
	}
	close(fd);
	return 0;
}

static int
decompose(s)
	u_char *s;
{
	int len, i, unit;

	len = strlen(s);
	for (i = len - 2; i >= 0; i --)
	{
		if (s[i] < '0' || s[i] > '9')
			break;
	}

	unit = atoi(&s[i + 1]);
	s[i + 1] = 0;
	return unit;
}

static void
usage(void)
{

	(void)fprintf(stderr, "%s%s",
	    "usage: vpdconfig [-dhv] special-file special-file slice-number\n",
	    "   ex: vpdconfig vpd0 wd0 0\n");
	exit(1);
}

static int
show_info(name, fd)
	char *name;
	int fd;
{
	struct vpd_unit_info vi;
	int error, slno;

	error = ioctl(fd, VPDIOGINFO, &vi);
	if (error < 0)
	{
		perror(name);
		return EIO;
	}

	slno = vi.vpd_slno;
	printf("%s:\t%s%ds%d\t%s\tstart\t%ld\tsize\t%ld\n",
		name, vi.vpd_xname, vi.vpd_unit, 
		slno, fstypenames[vi.vpd_cd.cd_se[slno].se_fstype],
		vi.vpd_cd.cd_se[slno].se_start,
		vi.vpd_cd.cd_se[slno].se_size);
	return 0;
}

